# stdlib
from collections.abc import Callable
from enum import Enum
import logging
import secrets
from typing import Any
from typing import cast

# third party
from result import Result

# relative
from ...abstract_server import ServerType
from ...client.client import HTTPConnection
from ...client.client import PythonConnection
from ...client.client import SyftClient
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...server.worker_settings import WorkerSettings
from ...service.settings.settings import ServerSettings
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...types.server_url import ServerURL
from ...types.transforms import TransformContext
from ...types.transforms import keep
from ...types.transforms import make_set_default
from ...types.transforms import transform
from ...types.transforms import transform_method
from ...types.uid import UID
from ...util.telemetry import instrument
from ...util.util import generate_token
from ...util.util import get_env
from ...util.util import prompt_warning_message
from ...util.util import str_to_bool
from ..context import AuthedServiceContext
from ..data_subject.data_subject import NamePartitionKey
from ..metadata.server_metadata import ServerMetadata
from ..request.request import Request
from ..request.request import RequestStatus
from ..request.request import SubmitRequest
from ..request.request_service import RequestService
from ..response import SyftError
from ..response import SyftInfo
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..warnings import CRUDWarning
from .association_request import AssociationRequestChange
from .reverse_tunnel_service import ReverseTunnelService
from .routes import HTTPServerRoute
from .routes import PythonServerRoute
from .routes import ServerRoute
from .routes import ServerRouteType
from .server_peer import ServerPeer
from .server_peer import ServerPeerUpdate

logger = logging.getLogger(__name__)

VerifyKeyPartitionKey = PartitionKey(key="verify_key", type_=SyftVerifyKey)
ServerTypePartitionKey = PartitionKey(key="server_type", type_=ServerType)
OrderByNamePartitionKey = PartitionKey(key="name", type_=str)

REVERSE_TUNNEL_ENABLED = "REVERSE_TUNNEL_ENABLED"


def reverse_tunnel_enabled() -> bool:
    return str_to_bool(get_env(REVERSE_TUNNEL_ENABLED, "false"))


@serializable(canonical_name="ServerPeerAssociationStatus", version=1)
class ServerPeerAssociationStatus(Enum):
    PEER_ASSOCIATED = "PEER_ASSOCIATED"
    PEER_ASSOCIATION_PENDING = "PEER_ASSOCIATION_PENDING"
    PEER_NOT_FOUND = "PEER_NOT_FOUND"


@instrument
@serializable(canonical_name="NetworkStash", version=1)
class NetworkStash(BaseUIDStoreStash):
    object_type = ServerPeer
    settings: PartitionSettings = PartitionSettings(
        name=ServerPeer.__canonical_name__, object_type=ServerPeer
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def get_by_name(
        self, credentials: SyftVerifyKey, name: str
    ) -> Result[ServerPeer | None, str]:
        qks = QueryKeys(qks=[NamePartitionKey.with_obj(name)])
        return self.query_one(credentials=credentials, qks=qks)

    def update(
        self,
        credentials: SyftVerifyKey,
        peer_update: ServerPeerUpdate,
        has_permission: bool = False,
    ) -> Result[ServerPeer, str]:
        valid = self.check_type(peer_update, ServerPeerUpdate)
        if valid.is_err():
            return SyftError(message=valid.err())
        return super().update(credentials, peer_update, has_permission=has_permission)

    def create_or_update_peer(
        self, credentials: SyftVerifyKey, peer: ServerPeer
    ) -> Result[ServerPeer, str]:
        """
        Update the selected peer and its route priorities if the peer already exists
        If the peer does not exist, simply adds it to the database.

        Args:
            credentials (SyftVerifyKey): The credentials used to authenticate the request.
            peer (ServerPeer): The peer to be updated or added.

        Returns:
            Result[ServerPeer, str]: The updated or added peer if the operation
            was successful, or an error message if the operation failed.
        """
        valid = self.check_type(peer, ServerPeer)
        if valid.is_err():
            return SyftError(message=valid.err())
        existing: Result | ServerPeer = self.get_by_uid(
            credentials=credentials, uid=peer.id
        )
        if existing.is_ok() and existing.ok():
            existing_peer = existing.ok()
            existing_peer.update_routes(peer.server_routes)
            peer_update = ServerPeerUpdate(
                id=peer.id, server_routes=existing_peer.server_routes
            )
            result = self.update(credentials, peer_update)
        else:
            result = self.set(credentials, peer)
        return result

    def get_by_verify_key(
        self, credentials: SyftVerifyKey, verify_key: SyftVerifyKey
    ) -> Result[ServerPeer | None, SyftError]:
        qks = QueryKeys(qks=[VerifyKeyPartitionKey.with_obj(verify_key)])
        return self.query_one(credentials, qks)

    def get_by_server_type(
        self, credentials: SyftVerifyKey, server_type: ServerType
    ) -> Result[list[ServerPeer], SyftError]:
        qks = QueryKeys(qks=[ServerTypePartitionKey.with_obj(server_type)])
        return self.query_all(
            credentials=credentials, qks=qks, order_by=OrderByNamePartitionKey
        )


@instrument
@serializable(canonical_name="NetworkService", version=1)
class NetworkService(AbstractService):
    store: DocumentStore
    stash: NetworkStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = NetworkStash(store=store)
        if reverse_tunnel_enabled():
            self.rtunnel_service = ReverseTunnelService()

    @service_method(
        path="network.exchange_credentials_with",
        name="exchange_credentials_with",
        roles=GUEST_ROLE_LEVEL,
        warning=CRUDWarning(confirmation=True),
    )
    def exchange_credentials_with(
        self,
        context: AuthedServiceContext,
        self_server_route: ServerRoute,
        remote_server_route: ServerRoute,
        remote_server_verify_key: SyftVerifyKey,
        reverse_tunnel: bool = False,
    ) -> Request | SyftSuccess | SyftError:
        """
        Exchange Route With Another Server. If there is a pending association request, return it
        """

        # Step 1: Validate the Route
        self_server_peer = self_server_route.validate_with_context(context=context)

        if reverse_tunnel and not reverse_tunnel_enabled():
            return SyftError(message="Reverse tunneling is not enabled on this server.")
        elif reverse_tunnel:
            _rtunnel_route = self_server_peer.server_routes[-1]
            _rtunnel_route.rtunnel_token = generate_token()
            _rtunnel_route.host_or_ip = f"{self_server_peer.name}.syft.local"
            self_server_peer.server_routes[-1] = _rtunnel_route

        if isinstance(self_server_peer, SyftError):
            return self_server_peer

        # Step 2: Send the Server Peer to the remote server
        # Also give them their own to validate that it belongs to them
        # random challenge prevents replay attacks
        remote_client: SyftClient = remote_server_route.client_with_context(
            context=context
        )
        remote_server_peer = ServerPeer.from_client(remote_client)

        # Step 3: Check remotely if the self server already exists as a peer
        # Update the peer if it exists, otherwise add it
        remote_self_server_peer = remote_client.api.services.network.get_peer_by_name(
            name=self_server_peer.name
        )

        association_request_approved = True
        if isinstance(remote_self_server_peer, ServerPeer):
            updated_peer = ServerPeerUpdate(
                id=self_server_peer.id, server_routes=self_server_peer.server_routes
            )
            result = remote_client.api.services.network.update_peer(
                peer_update=updated_peer
            )
            if isinstance(result, SyftError):
                logger.error(
                    f"Failed to update peer information on remote client. {result.message}"
                )
                return SyftError(
                    message=f"Failed to add peer information on remote client : {remote_client.id}"
                )

        # If  peer does not exist, ask the remote client to add this server
        # (represented by `self_server_peer`) as a peer
        if remote_self_server_peer is None:
            random_challenge = secrets.token_bytes(16)
            remote_res = remote_client.api.services.network.add_peer(
                peer=self_server_peer,
                challenge=random_challenge,
                self_server_route=remote_server_route,
                verify_key=remote_server_verify_key,
            )

            if isinstance(remote_res, SyftError):
                return SyftError(
                    message=f"Failed to add peer to remote client: {remote_client.id}. Error: {remote_res.message}"
                )

            association_request_approved = not isinstance(remote_res, Request)

        # Step 4: Save the remote peer for later
        result = self.stash.create_or_update_peer(
            context.server.verify_key,
            remote_server_peer,
        )
        if result.is_err():
            logging.error(
                f"Failed to save peer: {remote_server_peer}. Error: {result.err()}"
            )
            return SyftError(message="Failed to update route information.")

        # Step 5: Save config to enable reverse tunneling
        if reverse_tunnel and reverse_tunnel_enabled():
            self.set_reverse_tunnel_config(
                context=context,
                self_server_peer=self_server_peer,
                remote_server_peer=remote_server_peer,
            )

        return (
            SyftSuccess(message="Routes Exchanged")
            if association_request_approved
            else remote_res
        )

    @service_method(path="network.add_peer", name="add_peer", roles=GUEST_ROLE_LEVEL)
    def add_peer(
        self,
        context: AuthedServiceContext,
        peer: ServerPeer,
        challenge: bytes,
        self_server_route: ServerRoute,
        verify_key: SyftVerifyKey,
    ) -> Request | SyftSuccess | SyftError:
        """Add a Network Server Peer. Called by a remote server to add
        itself as a peer for the current server.
        """
        # Using the verify_key of the peer to verify the signature
        # It is also our single source of truth for the peer
        if peer.verify_key != context.credentials:
            return SyftError(
                message=(
                    f"The {type(peer).__name__}.verify_key: "
                    f"{peer.verify_key} does not match the signature of the message"
                )
            )

        if verify_key != context.server.verify_key:
            return SyftError(
                message="verify_key does not match the remote server's verify_key for add_peer"
            )

        # check if the peer already is a server peer
        existing_peer_res = self.stash.get_by_uid(context.server.verify_key, peer.id)
        if existing_peer_res.is_err():
            return SyftError(
                message=f"Failed to query peer from stash: {existing_peer_res.err()}"
            )

        if isinstance(existing_peer := existing_peer_res.ok(), ServerPeer):
            msg = [
                f"The peer '{peer.name}' is already associated with '{context.server.name}'"
            ]

            if existing_peer != peer:
                result = self.stash.create_or_update_peer(
                    context.server.verify_key,
                    peer,
                )
                msg.append("Peer information change detected.")

                if result.is_err():
                    msg.append("Attempt to update peer information failed.")
                    return SyftError(message="\n".join(msg))

                msg.append("Peer information successfully updated.")
                return SyftSuccess(message="\n".join(msg))

            return SyftSuccess(message="\n".join(msg))

        # check if the peer already submitted an association request
        association_requests: list[Request] = self._get_association_requests_by_peer_id(
            context=context, peer_id=peer.id
        )
        if (
            association_requests
            and (association_request := association_requests[-1]).status
            == RequestStatus.PENDING
        ):
            return association_request
        # only create and submit a new request if there is no requests yet
        # or all previous requests have been rejected
        association_request_change = AssociationRequestChange(
            self_server_route=self_server_route, challenge=challenge, remote_peer=peer
        )
        submit_request = SubmitRequest(
            changes=[association_request_change],
            requesting_user_verify_key=context.credentials,
        )
        request_submit_method = context.server.get_service_method(RequestService.submit)
        request = request_submit_method(context, submit_request)
        if (
            isinstance(request, Request)
            and context.server.settings.association_request_auto_approval
        ):
            request_apply_method = context.server.get_service_method(
                RequestService.apply
            )
            return request_apply_method(context, uid=request.id)

        return request

    @service_method(path="network.ping", name="ping", roles=GUEST_ROLE_LEVEL)
    def ping(
        self, context: AuthedServiceContext, challenge: bytes
    ) -> bytes | SyftError:
        """To check alivesness/authenticity of a peer"""

        # # Only the root user can ping the server to check its state
        # if context.server.verify_key != context.credentials:
        #     return SyftError(message=("Only the root user can access ping endpoint"))

        # this way they can match up who we are with who they think we are
        # Sending a signed messages for the peer to verify

        challenge_signature = context.server.signing_key.signing_key.sign(
            challenge
        ).signature

        return challenge_signature

    @service_method(
        path="network.check_peer_association",
        name="check_peer_association",
        roles=GUEST_ROLE_LEVEL,
    )
    def check_peer_association(
        self, context: AuthedServiceContext, peer_id: UID
    ) -> ServerPeerAssociationStatus | SyftError:
        """Check if a peer exists in the network stash"""

        # get the server peer for the given sender peer_id
        peer = self.stash.get_by_uid(context.server.verify_key, peer_id)
        if err := peer.is_err():
            return SyftError(message=f"Failed to query peer from stash. Err: {err}")

        if isinstance(peer.ok(), ServerPeer):
            return ServerPeerAssociationStatus.PEER_ASSOCIATED

        if peer.ok() is None:  # peer is either pending or not found
            association_requests: list[Request] = (
                self._get_association_requests_by_peer_id(
                    context=context, peer_id=peer_id
                )
            )
            if (
                association_requests
                and association_requests[-1].status == RequestStatus.PENDING
            ):
                return ServerPeerAssociationStatus.PEER_ASSOCIATION_PENDING

        return ServerPeerAssociationStatus.PEER_NOT_FOUND

    @service_method(
        path="network.get_all_peers", name="get_all_peers", roles=GUEST_ROLE_LEVEL
    )
    def get_all_peers(
        self, context: AuthedServiceContext
    ) -> list[ServerPeer] | SyftError:
        """Get all Peers"""

        result = self.stash.get_all(
            credentials=context.server.verify_key,
            order_by=OrderByNamePartitionKey,
        )
        if result.is_ok():
            peers = result.ok()
            return peers
        return SyftError(message=result.err())

    @service_method(
        path="network.get_peer_by_name", name="get_peer_by_name", roles=GUEST_ROLE_LEVEL
    )
    def get_peer_by_name(
        self, context: AuthedServiceContext, name: str
    ) -> ServerPeer | None | SyftError:
        """Get Peer by Name"""

        result = self.stash.get_by_name(
            credentials=context.server.verify_key,
            name=name,
        )
        if result.is_ok():
            peer = result.ok()
            return peer
        return SyftError(message=str(result.err()))

    @service_method(
        path="network.get_peers_by_type",
        name="get_peers_by_type",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_peers_by_type(
        self, context: AuthedServiceContext, server_type: ServerType
    ) -> list[ServerPeer] | SyftError:
        result = self.stash.get_by_server_type(
            credentials=context.server.verify_key,
            server_type=server_type,
        )

        if result.is_err():
            return SyftError(message=str(result.err()))

        # Return peers or an empty list when result is None
        return result.ok() or []

    @service_method(
        path="network.update_peer",
        name="update_peer",
        roles=GUEST_ROLE_LEVEL,
    )
    def update_peer(
        self,
        context: AuthedServiceContext,
        peer_update: ServerPeerUpdate,
    ) -> SyftSuccess | SyftError:
        # try setting all fields of ServerPeerUpdate according to ServerPeer

        result = self.stash.update(
            credentials=context.server.verify_key,
            peer_update=peer_update,
        )
        if result.is_err():
            return SyftError(
                message=f"Failed to update peer '{peer_update.name}'. Error: {result.err()}"
            )

        peer = result.ok()

        self.set_reverse_tunnel_config(context=context, remote_server_peer=peer)
        return SyftSuccess(
            message=f"Peer '{result.ok().name}' information successfully updated."
        )

    def set_reverse_tunnel_config(
        self,
        context: AuthedServiceContext,
        remote_server_peer: ServerPeer,
        self_server_peer: ServerPeer | None = None,
    ) -> None:
        server_type = cast(ServerType, context.server.server_type)
        if server_type.value == ServerType.GATEWAY.value:
            rtunnel_route = remote_server_peer.get_rtunnel_route()
            (
                self.rtunnel_service.set_server_config(remote_server_peer)
                if rtunnel_route
                else None
            )
        else:
            self_server_peer = (
                context.server.settings.to(ServerPeer)
                if self_server_peer is None
                else self_server_peer
            )
            rtunnel_route = self_server_peer.get_rtunnel_route()
            (
                self.rtunnel_service.set_client_config(
                    self_server_peer=self_server_peer,
                    remote_server_route=remote_server_peer.pick_highest_priority_route(),
                )
                if rtunnel_route
                else None
            )

    @service_method(
        path="network.delete_peer_by_id",
        name="delete_peer_by_id",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def delete_peer_by_id(
        self, context: AuthedServiceContext, uid: UID
    ) -> SyftSuccess | SyftError:
        """Delete Server Peer"""
        retrieve_result = self.stash.get_by_uid(context.credentials, uid)
        if err := retrieve_result.is_err():
            return SyftError(
                message=f"Failed to retrieve peer with UID {uid}: {retrieve_result.err()}."
            )
        peer_to_delete = cast(ServerPeer, retrieve_result.ok())

        server_side_type = cast(ServerType, context.server.server_type)
        if server_side_type.value == ServerType.GATEWAY.value:
            rtunnel_route = peer_to_delete.get_rtunnel_route()
            (
                self.rtunnel_service.clear_server_config(peer_to_delete)
                if rtunnel_route
                else None
            )

        # TODO: Handle the case when peer is deleted from datasite server

        result = self.stash.delete_by_uid(context.credentials, uid)
        if err := result.is_err():
            return SyftError(message=f"Failed to delete peer with UID {uid}: {err}.")
        # Delete all the association requests from this peer
        association_requests: list[Request] = self._get_association_requests_by_peer_id(
            context=context, peer_id=uid
        )
        for request in association_requests:
            request_delete_method = context.server.get_service_method(
                RequestService.delete_by_uid
            )
            res = request_delete_method(context, request.id)
            if isinstance(res, SyftError):
                return res
        # TODO: Notify the peer (either by email or by other form of notifications)
        # that it has been deleted from the network
        return SyftSuccess(message=f"Server Peer with id {uid} deleted.")

    @service_method(path="network.add_route_on_peer", name="add_route_on_peer")
    def add_route_on_peer(
        self,
        context: AuthedServiceContext,
        peer: ServerPeer,
        route: ServerRoute,
    ) -> SyftSuccess | SyftError:
        """
        Add or update the route information on the remote peer.

        Args:
            context (AuthedServiceContext): The authentication context.
            peer (ServerPeer): The peer representing the remote server.
            route (ServerRoute): The route to be added.

        Returns:
            SyftSuccess | SyftError: A success message if the route is verified,
                otherwise an error message.
        """
        # creates a client on the remote server based on the credentials
        # of the current server's client
        remote_client = peer.client_with_context(context=context)
        if remote_client.is_err():
            return SyftError(
                message=f"Failed to create remote client for peer: "
                f"{peer.id}. Error: {remote_client.err()}"
            )
        remote_client = remote_client.ok()
        # ask the remote server to add the route to the self server
        result = remote_client.api.services.network.add_route(
            peer_verify_key=context.credentials,
            route=route,
            called_by_peer=True,
        )
        return result

    @service_method(path="network.add_route", name="add_route", roles=GUEST_ROLE_LEVEL)
    def add_route(
        self,
        context: AuthedServiceContext,
        peer_verify_key: SyftVerifyKey,
        route: ServerRoute,
        called_by_peer: bool = False,
    ) -> SyftSuccess | SyftError:
        """
        Add a route to the peer. If the route already exists, update its priority.

        Args:
            context (AuthedServiceContext): The authentication context of the remote server.
            peer_verify_key (SyftVerifyKey): The verify key of the remote server peer.
            route (ServerRoute): The route to be added.
            called_by_peer (bool): The flag to indicate that it's called by a remote peer.

        Returns:
            SyftSuccess | SyftError
        """
        # verify if the peer is truly the one sending the request to add the route to itself
        if called_by_peer and peer_verify_key != context.credentials:
            return SyftError(
                message=(
                    f"The {type(peer_verify_key).__name__}: "
                    f"{peer_verify_key} does not match the signature of the message"
                )
            )
        # get the full peer object from the store to update its routes
        remote_server_peer: ServerPeer | SyftError = (
            self._get_remote_server_peer_by_verify_key(context, peer_verify_key)
        )
        if isinstance(remote_server_peer, SyftError):
            return remote_server_peer
        # add and update the priority for the peer
        if route in remote_server_peer.server_routes:
            return SyftSuccess(
                message=f"The route already exists between '{context.server.name}' and "
                f"peer '{remote_server_peer.name}'."
            )

        remote_server_peer.update_route(route=route)
        # update the peer in the store with the updated routes
        peer_update = ServerPeerUpdate(
            id=remote_server_peer.id, server_routes=remote_server_peer.server_routes
        )
        result = self.stash.update(
            credentials=context.server.verify_key,
            peer_update=peer_update,
        )
        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(
            message=f"New route ({str(route)}) with id '{route.id}' "
            f"to peer {remote_server_peer.server_type.value} '{remote_server_peer.name}' "
            f"was added for {str(context.server.server_type)} '{context.server.name}'"
        )

    @service_method(path="network.delete_route_on_peer", name="delete_route_on_peer")
    def delete_route_on_peer(
        self,
        context: AuthedServiceContext,
        peer: ServerPeer,
        route: ServerRoute,
    ) -> SyftSuccess | SyftError | SyftInfo:
        """
        Delete the route on the remote peer.

        Args:
            context (AuthedServiceContext): The authentication context for the service.
            peer (ServerPeer): The peer for which the route will be deleted.
            route (ServerRoute): The route to be deleted.

        Returns:
            SyftSuccess: If the route is successfully deleted.
            SyftError: If there is an error deleting the route.
            SyftInfo: If there is only one route left for the peer and
                the admin chose not to remove it
        """
        # creates a client on the remote server based on the credentials
        # of the current server's client
        remote_client = peer.client_with_context(context=context)
        if remote_client.is_err():
            return SyftError(
                message=f"Failed to create remote client for peer: "
                f"{peer.id}. Error: {remote_client.err()}"
            )
        remote_client = remote_client.ok()
        # ask the remote server to delete the route to the self server,
        result = remote_client.api.services.network.delete_route(
            peer_verify_key=context.credentials,
            route=route,
            called_by_peer=True,
        )
        return result

    @service_method(
        path="network.delete_route", name="delete_route", roles=GUEST_ROLE_LEVEL
    )
    def delete_route(
        self,
        context: AuthedServiceContext,
        peer_verify_key: SyftVerifyKey,
        route: ServerRoute | None = None,
        called_by_peer: bool = False,
    ) -> SyftSuccess | SyftError | SyftInfo:
        """
        Delete a route for a given peer.
        If a peer has no routes left, there will be a prompt asking if the user want to remove it.
        If the answer is yes, it will be removed from the stash and will no longer be a peer.

        Args:
            context (AuthedServiceContext): The authentication context for the service.
            peer_verify_key (SyftVerifyKey): The verify key of the remote server peer.
            route (ServerRoute): The route to be deleted.
            called_by_peer (bool): The flag to indicate that it's called by a remote peer.

        Returns:
            SyftSuccess: If the route is successfully deleted.
            SyftError: If there is an error deleting the route.
            SyftInfo: If there is only one route left for the peer and
                the admin chose not to remove it
        """
        if called_by_peer and peer_verify_key != context.credentials:
            # verify if the peer is truly the one sending the request to delete the route to itself
            return SyftError(
                message=(
                    f"The {type(peer_verify_key).__name__}: "
                    f"{peer_verify_key} does not match the signature of the message"
                )
            )

        remote_server_peer: ServerPeer | SyftError = (
            self._get_remote_server_peer_by_verify_key(
                context=context, peer_verify_key=peer_verify_key
            )
        )

        if len(remote_server_peer.server_routes) == 1:
            warning_message = (
                f"There is only one route left to peer "
                f"{remote_server_peer.server_type.value} '{remote_server_peer.name}'. "
                f"Removing this route will remove the peer for "
                f"{str(context.server.server_type)} '{context.server.name}'."
            )
            response: bool = prompt_warning_message(
                message=warning_message,
                confirm=False,
            )
            if not response:
                return SyftInfo(
                    message=f"The last route to {remote_server_peer.server_type.value} "
                    f"'{remote_server_peer.name}' with id "
                    f"'{remote_server_peer.server_routes[0].id}' was not deleted."
                )

        result = remote_server_peer.delete_route(route=route)
        return_message = (
            f"Route '{str(route)}' to peer "
            f"{remote_server_peer.server_type.value} '{remote_server_peer.name}' "
            f"was deleted for {str(context.server.server_type)} '{context.server.name}'."
        )
        if isinstance(result, SyftError):
            return result

        if len(remote_server_peer.server_routes) == 0:
            # remove the peer
            # TODO: should we do this as we are deleting the peer with a guest role level?
            result = self.stash.delete_by_uid(
                credentials=context.server.verify_key, uid=remote_server_peer.id
            )
            if isinstance(result, SyftError):
                return result
            return_message += (
                f" There is no routes left to connect to peer "
                f"{remote_server_peer.server_type.value} '{remote_server_peer.name}', so it is deleted for "
                f"{str(context.server.server_type)} '{context.server.name}'."
            )
        else:
            # update the peer with the route removed
            peer_update = ServerPeerUpdate(
                id=remote_server_peer.id, server_routes=remote_server_peer.server_routes
            )
            result = self.stash.update(
                credentials=context.server.verify_key, peer_update=peer_update
            )
            if result.is_err():
                return SyftError(message=str(result.err()))

        return SyftSuccess(message=return_message)

    @service_method(
        path="network.update_route_priority_on_peer",
        name="update_route_priority_on_peer",
    )
    def update_route_priority_on_peer(
        self,
        context: AuthedServiceContext,
        peer: ServerPeer,
        route: ServerRoute,
        priority: int | None = None,
    ) -> SyftSuccess | SyftError:
        """
        Update the route priority on the remote peer.

        Args:
            context (AuthedServiceContext): The authentication context.
            peer (ServerPeer): The peer representing the remote server.
            route (ServerRoute): The route to be added.
            priority (int | None): The new priority value for the route. If not
                provided, it will be assigned the highest priority among all peers

        Returns:
            SyftSuccess | SyftError: A success message if the route is verified,
                otherwise an error message.
        """
        # creates a client on the remote server based on the credentials
        # of the current server's client
        remote_client = peer.client_with_context(context=context)
        if remote_client.is_err():
            return SyftError(
                message=f"Failed to create remote client for peer: "
                f"{peer.id}. Error: {remote_client.err()}"
            )
        remote_client = remote_client.ok()
        result = remote_client.api.services.network.update_route_priority(
            peer_verify_key=context.credentials,
            route=route,
            priority=priority,
            called_by_peer=True,
        )
        return result

    @service_method(
        path="network.update_route_priority",
        name="update_route_priority",
        roles=GUEST_ROLE_LEVEL,
    )
    def update_route_priority(
        self,
        context: AuthedServiceContext,
        peer_verify_key: SyftVerifyKey,
        route: ServerRoute,
        priority: int | None = None,
        called_by_peer: bool = False,
    ) -> SyftSuccess | SyftError:
        """
        Updates a route's priority for the given peer

        Args:
            context (AuthedServiceContext): The authentication context for the service.
            peer_verify_key (SyftVerifyKey): The verify key of the peer whose route priority needs to be updated.
            route (ServerRoute): The route for which the priority needs to be updated.
            priority (int | None): The new priority value for the route. If not
                provided, it will be assigned the highest priority among all peers

        Returns:
            SyftSuccess | SyftError: Successful / Error response
        """
        if called_by_peer and peer_verify_key != context.credentials:
            return SyftError(
                message=(
                    f"The {type(peer_verify_key).__name__}: "
                    f"{peer_verify_key} does not match the signature of the message"
                )
            )
        # get the full peer object from the store to update its routes
        remote_server_peer: ServerPeer | SyftError = (
            self._get_remote_server_peer_by_verify_key(context, peer_verify_key)
        )
        if isinstance(remote_server_peer, SyftError):
            return remote_server_peer
        # update the route's priority for the peer
        updated_server_route: ServerRouteType | SyftError = (
            remote_server_peer.update_existed_route_priority(
                route=route, priority=priority
            )
        )
        if isinstance(updated_server_route, SyftError):
            return updated_server_route
        new_priority: int = updated_server_route.priority
        # update the peer in the store
        peer_update = ServerPeerUpdate(
            id=remote_server_peer.id, server_routes=remote_server_peer.server_routes
        )
        result = self.stash.update(context.server.verify_key, peer_update)
        if result.is_err():
            return SyftError(message=str(result.err()))

        return SyftSuccess(
            message=f"Route {route.id}'s priority updated to "
            f"{new_priority} for peer {remote_server_peer.name}"
        )

    def _get_remote_server_peer_by_verify_key(
        self, context: AuthedServiceContext, peer_verify_key: SyftVerifyKey
    ) -> ServerPeer | SyftError:
        """
        Helper function to get the full server peer object from t
        he stash using its verify key
        """
        remote_server_peer: Result[ServerPeer | None, SyftError] = (
            self.stash.get_by_verify_key(
                credentials=context.server.verify_key,
                verify_key=peer_verify_key,
            )
        )
        if remote_server_peer.is_err():
            return SyftError(message=str(remote_server_peer.err()))
        remote_server_peer = remote_server_peer.ok()
        if remote_server_peer is None:
            return SyftError(
                message=f"Can't retrieve {remote_server_peer.name} from the store of peers (None)."
            )
        return remote_server_peer

    def _get_association_requests_by_peer_id(
        self, context: AuthedServiceContext, peer_id: UID
    ) -> list[Request]:
        """
        Get all the association requests from a peer. The association requests are sorted by request_time.
        """
        request_get_all_method: Callable = context.server.get_service_method(
            RequestService.get_all
        )
        all_requests: list[Request] = request_get_all_method(context)
        association_requests: list[Request] = [
            request
            for request in all_requests
            if any(
                isinstance(change, AssociationRequestChange)
                and change.remote_peer.id == peer_id
                for change in request.changes
            )
        ]

        return sorted(
            association_requests, key=lambda request: request.request_time.utc_timestamp
        )


TYPE_TO_SERVICE[ServerPeer] = NetworkService
SERVICE_TO_TYPES[NetworkService].update({ServerPeer})


def from_server_url(context: TransformContext) -> TransformContext:
    if context.obj is not None and context.output is not None:
        url = context.obj.url.as_container_host()
        context.output["host_or_ip"] = url.host_or_ip
        context.output["protocol"] = url.protocol
        context.output["port"] = url.port
        context.output["private"] = False
        context.output["proxy_target_uid"] = context.obj.proxy_target_uid
        context.output["priority"] = 1
        context.output["rtunnel_token"] = context.obj.rtunnel_token

    return context


@transform(HTTPConnection, HTTPServerRoute)
def http_connection_to_server_route() -> list[Callable]:
    return [from_server_url]


def get_python_server_route(context: TransformContext) -> TransformContext:
    if context.output is not None and context.obj is not None:
        context.output["id"] = context.obj.server.id
        context.output["worker_settings"] = WorkerSettings.from_server(
            context.obj.server
        )
        context.output["proxy_target_uid"] = context.obj.proxy_target_uid
    return context


@transform(PythonConnection, PythonServerRoute)
def python_connection_to_server_route() -> list[Callable]:
    return [get_python_server_route]


@transform_method(PythonServerRoute, PythonConnection)
def server_route_to_python_connection(
    obj: Any, context: TransformContext | None = None
) -> list[Callable]:
    return PythonConnection(server=obj.server, proxy_target_uid=obj.proxy_target_uid)


@transform_method(HTTPServerRoute, HTTPConnection)
def server_route_to_http_connection(
    obj: Any, context: TransformContext | None = None
) -> list[Callable]:
    url = ServerURL(
        protocol=obj.protocol, host_or_ip=obj.host_or_ip, port=obj.port
    ).as_container_host()
    return HTTPConnection(
        url=url,
        proxy_target_uid=obj.proxy_target_uid,
        rtunnel_token=obj.rtunnel_token,
    )


@transform(ServerMetadata, ServerPeer)
def metadata_to_peer() -> list[Callable]:
    return [
        keep(
            [
                "id",
                "name",
                "verify_key",
                "server_type",
            ]
        ),
        make_set_default("admin_email", ""),
    ]


@transform(ServerSettings, ServerPeer)
def settings_to_peer() -> list[Callable]:
    return [
        keep(["id", "name", "verify_key", "server_type", "admin_email"]),
    ]



# stdlib

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionSettings
from .api import TwinAPIEndpoint

MISSING_PATH_STRING = "Endpoint path: {path} does not exist."


@serializable(canonical_name="TwinAPIEndpointStash", version=1)
class TwinAPIEndpointStash(BaseUIDStoreStash):
    object_type = TwinAPIEndpoint
    settings: PartitionSettings = PartitionSettings(
        name=TwinAPIEndpoint.__canonical_name__, object_type=TwinAPIEndpoint
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def get_by_path(
        self, credentials: SyftVerifyKey, path: str
    ) -> Result[TwinAPIEndpoint, str]:
        endpoint_results = self.get_all(credentials=credentials)
        if endpoint_results.is_err():
            return endpoint_results

        endpoints = []
        if endpoint_results.is_ok():
            endpoints = endpoint_results.ok()

        for endpoint in endpoints:
            if endpoint.path == path:
                return Ok(endpoint)

        return Err(MISSING_PATH_STRING.format(path=path))

    def path_exists(self, credentials: SyftVerifyKey, path: str) -> Result[bool, str]:
        result = self.get_by_path(credentials=credentials, path=path)

        if result.is_err() and result.err() == MISSING_PATH_STRING.format(path=path):
            return Ok(False)
        if result.is_ok():
            return Ok(True)

        return Err(result.err())

    def upsert(
        self,
        credentials: SyftVerifyKey,
        endpoint: TwinAPIEndpoint,
        has_permission: bool = False,
    ) -> Result[TwinAPIEndpoint, str]:
        """Upsert an endpoint."""
        result = self.path_exists(credentials=credentials, path=endpoint.path)

        if result.is_err():
            return result

        if result.ok():
            super().delete_by_uid(credentials=credentials, uid=endpoint.id)

        result = super().set(
            credentials=credentials, obj=endpoint, ignore_duplicates=False
        )
        return result



# stdlib
import importlib
import logging
from typing import Any
from typing import cast

# third party
import numpy as np
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...types.datetime import DateTime
from ...types.syft_object import SyftObject
from ...types.twin_object import TwinObject
from ...types.uid import UID
from ..blob_storage.service import BlobStorageService
from ..code.user_code import UserCode
from ..code.user_code import execute_byte_code
from ..context import AuthedServiceContext
from ..policy.policy import OutputPolicy
from ..policy.policy import retrieve_from_db
from ..response import SyftError
from ..response import SyftSuccess
from ..response import SyftWarning
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from ..service import UserLibConfigRegistry
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from ..user.user_roles import ServiceRole
from .action_endpoint import CustomEndpointActionObject
from .action_object import Action
from .action_object import ActionObject
from .action_object import ActionObjectPointer
from .action_object import ActionType
from .action_object import AnyActionObject
from .action_object import TwinMode
from .action_permissions import ActionObjectPermission
from .action_permissions import ActionObjectREAD
from .action_permissions import ActionPermission
from .action_store import ActionStore
from .action_types import action_type_for_type
from .numpy import NumpyArrayObject
from .pandas import PandasDataFrameObject  # noqa: F401
from .pandas import PandasSeriesObject  # noqa: F401

logger = logging.getLogger(__name__)


@serializable(canonical_name="ActionService", version=1)
class ActionService(AbstractService):
    store_type = ActionStore

    def __init__(self, store: ActionStore) -> None:
        self.store = store

    @service_method(path="action.np_array", name="np_array")
    def np_array(self, context: AuthedServiceContext, data: Any) -> Any:
        if not isinstance(data, np.ndarray):
            data = np.array(data)
        # cast here since we are sure that AuthedServiceContext has a server

        np_obj = NumpyArrayObject(
            dtype=data.dtype,
            shape=data.shape,
            syft_action_data_cache=data,
            syft_server_location=context.server.id,
            syft_client_verify_key=context.credentials,
        )
        blob_store_result = np_obj._save_to_blob_storage()
        if isinstance(blob_store_result, SyftError):
            return blob_store_result
        if isinstance(blob_store_result, SyftWarning):
            logger.debug(blob_store_result.message)

        np_pointer = self._set(context, np_obj)
        return np_pointer

    @service_method(
        path="action.set",
        name="set",
        roles=GUEST_ROLE_LEVEL,
    )
    def set(
        self,
        context: AuthedServiceContext,
        action_object: ActionObject | TwinObject,
        add_storage_permission: bool = True,
        ignore_detached_objs: bool = False,
    ) -> ActionObject | SyftError:
        res = self._set(
            context,
            action_object,
            has_result_read_permission=True,
            add_storage_permission=add_storage_permission,
            ignore_detached_objs=ignore_detached_objs,
        )
        if res.is_err():
            return SyftError(message=res.value)
        else:
            return res.ok()

    def is_detached_obj(
        self,
        action_object: ActionObject | TwinObject,
        ignore_detached_obj: bool = False,
    ) -> bool:
        """
        A detached object is an object that is not yet saved to the blob storage.
        """
        if (
            isinstance(action_object, TwinObject)
            and (
                (
                    action_object.mock_obj.syft_action_saved_to_blob_store
                    and action_object.mock_obj.syft_blob_storage_entry_id is None
                )
                or (
                    action_object.private_obj.syft_action_saved_to_blob_store
                    and action_object.private_obj.syft_blob_storage_entry_id is None
                )
            )
            and not ignore_detached_obj
        ):
            return True
        if isinstance(action_object, ActionObject) and (
            action_object.syft_action_saved_to_blob_store
            and action_object.syft_blob_storage_entry_id is None
            and not ignore_detached_obj
        ):
            return True
        return False

    def _set(
        self,
        context: AuthedServiceContext,
        action_object: ActionObject | TwinObject,
        has_result_read_permission: bool = False,
        add_storage_permission: bool = True,
        ignore_detached_objs: bool = False,
    ) -> Result[ActionObject, str]:
        if self.is_detached_obj(action_object, ignore_detached_objs):
            return Err(
                "You uploaded an ActionObject that is not yet in the blob storage"
            )
        """Save an object to the action store"""
        # 🟡 TODO 9: Create some kind of type checking / protocol for SyftSerializable

        if isinstance(action_object, ActionObject):
            action_object.syft_created_at = DateTime.now()
            (
                action_object._clear_cache()
                if action_object.syft_action_saved_to_blob_store
                else None
            )
        else:  # TwinObject
            action_object.private_obj.syft_created_at = DateTime.now()  # type: ignore[unreachable]
            action_object.mock_obj.syft_created_at = DateTime.now()

            # Clear cache if data is saved to blob storage
            (
                action_object.private_obj._clear_cache()
                if action_object.private_obj.syft_action_saved_to_blob_store
                else None
            )
            (
                action_object.mock_obj._clear_cache()
                if action_object.mock_obj.syft_action_saved_to_blob_store
                else None
            )

        # If either context or argument is True, has_result_read_permission is True
        has_result_read_permission = (
            context.extra_kwargs.get("has_result_read_permission", False)
            or has_result_read_permission
        )

        result = self.store.set(
            uid=action_object.id,
            credentials=context.credentials,
            syft_object=action_object,
            has_result_read_permission=has_result_read_permission,
            add_storage_permission=add_storage_permission,
        )
        if result.is_ok():
            if isinstance(action_object, TwinObject):
                # give read permission to the mock
                # if mock is saved to blob store, then add READ permission
                if action_object.mock_obj.syft_action_saved_to_blob_store:
                    blob_id = action_object.mock_obj.syft_blob_storage_entry_id
                    permission = ActionObjectPermission(
                        blob_id, ActionPermission.ALL_READ
                    )
                    blob_storage_service: AbstractService = context.server.get_service(
                        BlobStorageService
                    )
                    blob_storage_service.stash.add_permission(permission)
                if has_result_read_permission:
                    action_object = action_object.private
                else:
                    action_object = action_object.mock

            action_object.syft_point_to(context.server.id)
            return Ok(action_object)
        return result.err()

    @service_method(
        path="action.is_resolved", name="is_resolved", roles=GUEST_ROLE_LEVEL
    )
    def is_resolved(
        self,
        context: AuthedServiceContext,
        uid: UID,
    ) -> Result[Ok[bool], Err[str]]:
        """Get an object from the action store"""
        result = self._get(context, uid)
        if result.is_ok():
            obj = result.ok()
            if obj.is_link:
                result = self.resolve_links(
                    context, obj.syft_action_data.action_object_id.id
                )
                # Checking in case any error occurred
                if result.is_err():
                    return result

                return Ok(result.syft_resolved)

            # If it's a leaf but not resolved yet, return false
            elif not obj.syft_resolved:
                return Ok(False)

            # If it's not an action data link or non resolved (empty). It's resolved
            return Ok(True)
        # If it's not in the store or permission error, return the error
        return result

    @service_method(
        path="action.resolve_links", name="resolve_links", roles=GUEST_ROLE_LEVEL
    )
    def resolve_links(
        self,
        context: AuthedServiceContext,
        uid: UID,
        twin_mode: TwinMode = TwinMode.PRIVATE,
    ) -> Result[Ok[ActionObject], Err[str]]:
        """Get an object from the action store"""
        # relative

        result = self.store.get(uid=uid, credentials=context.credentials)
        # If user has permission to get the object / object exists
        if result.is_ok():
            obj = result.ok()

            # If it's not a leaf
            if obj.is_link:
                nested_result = self.resolve_links(
                    context, obj.syft_action_data.action_object_id.id, twin_mode
                )
                return nested_result

            # If it's a leaf
            return result

        return result

    @service_method(path="action.get", name="get", roles=GUEST_ROLE_LEVEL)
    def get(
        self,
        context: AuthedServiceContext,
        uid: UID,
        twin_mode: TwinMode = TwinMode.PRIVATE,
        resolve_nested: bool = True,
    ) -> Result[ActionObject, str]:
        """Get an object from the action store"""
        return self._get(context, uid, twin_mode, resolve_nested=resolve_nested)

    def _get(
        self,
        context: AuthedServiceContext,
        uid: UID,
        twin_mode: TwinMode = TwinMode.PRIVATE,
        has_permission: bool = False,
        resolve_nested: bool = True,
    ) -> Result[ActionObject, str]:
        """Get an object from the action store"""
        result = self.store.get(
            uid=uid, credentials=context.credentials, has_permission=has_permission
        )
        if result.is_ok() and context.server is not None:
            obj: TwinObject | ActionObject = result.ok()
            obj._set_obj_location_(
                context.server.id,
                context.credentials,
            )
            # Resolve graph links
            if (
                not isinstance(obj, TwinObject)  # type: ignore[unreachable]
                and resolve_nested
                and obj.is_link
            ):
                if not self.is_resolved(  # type: ignore[unreachable]
                    context, obj.syft_action_data.action_object_id.id
                ).ok():
                    return SyftError(message="This object is not resolved yet.")
                result = self.resolve_links(
                    context, obj.syft_action_data.action_object_id.id, twin_mode
                )
                return result
            if isinstance(obj, TwinObject):
                if twin_mode == TwinMode.PRIVATE:
                    obj = obj.private
                    obj.syft_point_to(context.server.id)
                elif twin_mode == TwinMode.MOCK:
                    obj = obj.mock
                    obj.syft_point_to(context.server.id)
                else:
                    obj.mock.syft_point_to(context.server.id)
                    obj.private.syft_point_to(context.server.id)
            return Ok(obj)
        else:
            return result

    @service_method(
        path="action.get_pointer", name="get_pointer", roles=GUEST_ROLE_LEVEL
    )
    def get_pointer(
        self, context: AuthedServiceContext, uid: UID
    ) -> Result[ActionObjectPointer, str]:
        """Get a pointer from the action store"""

        result = self.store.get_pointer(
            uid=uid, credentials=context.credentials, server_uid=context.server.id
        )
        if result.is_ok():
            obj = result.ok()
            obj._set_obj_location_(
                context.server.id,
                context.credentials,
            )
            return Ok(obj)
        return Err(result.err())

    @service_method(path="action.get_mock", name="get_mock", roles=GUEST_ROLE_LEVEL)
    def get_mock(
        self, context: AuthedServiceContext, uid: UID
    ) -> Result[SyftError, SyftObject]:
        """Get a pointer from the action store"""
        result = self.store.get_mock(uid=uid)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(
        path="action.has_storage_permission",
        name="has_storage_permission",
        roles=GUEST_ROLE_LEVEL,
    )
    def has_storage_permission(self, context: AuthedServiceContext, uid: UID) -> bool:
        return self.store.has_storage_permission(uid)

    def has_read_permission(self, context: AuthedServiceContext, uid: UID) -> bool:
        return self.store.has_permissions(
            [ActionObjectREAD(uid=uid, credentials=context.credentials)]
        )

    # not a public service endpoint
    def _user_code_execute(
        self,
        context: AuthedServiceContext,
        code_item: UserCode,
        kwargs: dict[str, Any],
        result_id: UID | None = None,
    ) -> Result[ActionObjectPointer, Err]:
        override_execution_permission = (
            context.has_execute_permissions or context.role == ServiceRole.ADMIN
        )
        if context.server:
            user_code_service = context.server.get_service("usercodeservice")

        input_policy = code_item.get_input_policy(context)
        output_policy = code_item.get_output_policy(context)

        # Unwrap nested ActionObjects
        for _k, arg in kwargs.items():
            self.flatten_action_arg(context, arg) if isinstance(arg, UID) else None

        if not override_execution_permission:
            if input_policy is None:
                if not code_item.is_output_policy_approved(context):
                    return Err("Execution denied: Your code is waiting for approval")
                return Err(f"No input policy defined for user code: {code_item.id}")

            # Filter input kwargs based on policy
            filtered_kwargs = input_policy.filter_kwargs(
                kwargs=kwargs, context=context, code_item_id=code_item.id
            )
            if filtered_kwargs.is_err():
                return filtered_kwargs
            filtered_kwargs = filtered_kwargs.ok()

            # validate input policy
            is_approved = input_policy._is_valid(
                context=context,
                usr_input_kwargs=kwargs,
                code_item_id=code_item.id,
            )
            if is_approved.is_err():
                return is_approved
        else:
            result = retrieve_from_db(code_item.id, kwargs, context)
            if isinstance(result, SyftError):
                return Err(result.message)
            filtered_kwargs = result.ok()

        if hasattr(input_policy, "transform_kwargs"):
            filtered_kwargs_res = input_policy.transform_kwargs(  # type: ignore
                context,
                filtered_kwargs,
            )
            if filtered_kwargs_res.is_err():
                return filtered_kwargs_res
            else:
                filtered_kwargs = filtered_kwargs_res.ok()

        # update input policy to track any input state

        has_twin_inputs = False

        real_kwargs = {}
        for key, kwarg_value in filtered_kwargs.items():
            if isinstance(kwarg_value, TwinObject):
                has_twin_inputs = True
            real_kwargs[key] = kwarg_value

        result_id = UID() if result_id is None else result_id

        try:
            if not has_twin_inputs:
                # no twins
                # allow python types from inputpolicy
                filtered_kwargs = filter_twin_kwargs(
                    real_kwargs, twin_mode=TwinMode.NONE, allow_python_types=True
                )
                exec_result = execute_byte_code(code_item, filtered_kwargs, context)
                if output_policy:
                    exec_result.result = output_policy.apply_to_output(
                        context,
                        exec_result.result,
                        update_policy=not override_execution_permission,
                    )
                code_item.output_policy = output_policy  # type: ignore
                user_code_service.update_code_state(context, code_item)
                if isinstance(exec_result.result, ActionObject):
                    result_action_object = ActionObject.link(
                        result_id=result_id, pointer_id=exec_result.result.id
                    )
                else:
                    result_action_object = wrap_result(result_id, exec_result.result)
            else:
                # twins
                private_kwargs = filter_twin_kwargs(
                    real_kwargs, twin_mode=TwinMode.PRIVATE, allow_python_types=True
                )
                private_exec_result = execute_byte_code(
                    code_item, private_kwargs, context
                )
                if output_policy:
                    private_exec_result.result = output_policy.apply_to_output(
                        context,
                        private_exec_result.result,
                        update_policy=not override_execution_permission,
                    )
                code_item.output_policy = output_policy  # type: ignore
                user_code_service.update_code_state(context, code_item)
                result_action_object_private = wrap_result(
                    result_id, private_exec_result.result
                )

                mock_kwargs = filter_twin_kwargs(
                    real_kwargs, twin_mode=TwinMode.MOCK, allow_python_types=True
                )
                # relative
                from .action_data_empty import ActionDataEmpty

                if any(isinstance(v, ActionDataEmpty) for v in mock_kwargs.values()):
                    mock_exec_result_obj = ActionDataEmpty()
                else:
                    mock_exec_result = execute_byte_code(
                        code_item, mock_kwargs, context
                    )
                    if output_policy:
                        mock_exec_result.result = output_policy.apply_to_output(
                            context, mock_exec_result.result, update_policy=False
                        )
                    mock_exec_result_obj = mock_exec_result.result

                result_action_object_mock = wrap_result(result_id, mock_exec_result_obj)

                result_action_object = TwinObject(
                    id=result_id,
                    private_obj=result_action_object_private,
                    mock_obj=result_action_object_mock,
                )
        except Exception as e:
            return Err(f"_user_code_execute failed. {e}")
        return Ok(result_action_object)

    def set_result_to_store(
        self,
        result_action_object: ActionObject | TwinObject,
        context: AuthedServiceContext,
        output_policy: OutputPolicy | None = None,
        has_result_read_permission: bool = False,
    ) -> Result[ActionObject, str]:
        result_id = result_action_object.id
        # result_blob_id = result_action_object.syft_blob_storage_entry_id

        if output_policy is not None:
            output_readers = (
                output_policy.output_readers
                if not context.has_execute_permissions
                else []
            )
        else:
            output_readers = []

        # If flag is True, user has read permissions to the results in BlobStore
        if has_result_read_permission:
            output_readers.append(context.credentials)

        read_permission = ActionPermission.READ

        result_action_object._set_obj_location_(
            context.server.id,
            context.credentials,
        )
        blob_store_result = result_action_object._save_to_blob_storage()
        if isinstance(blob_store_result, SyftError):
            return Err(blob_store_result.message)
        if isinstance(blob_store_result, SyftWarning):
            logger.debug(blob_store_result.message)

        # IMPORTANT: DO THIS ONLY AFTER ._save_to_blob_storage
        if isinstance(result_action_object, TwinObject):
            result_blob_id = result_action_object.private.syft_blob_storage_entry_id
        else:
            result_blob_id = result_action_object.syft_blob_storage_entry_id  # type: ignore[unreachable]

        # pass permission information to the action store as extra kwargs
        # context.extra_kwargs = {"has_result_read_permission": True}

        # Since this just meta data about the result, they always have access to it.
        set_result = self._set(
            context,
            result_action_object,
            has_result_read_permission=True,
        )

        if set_result.is_err():
            return set_result

        blob_storage_service: AbstractService = context.server.get_service(
            BlobStorageService
        )

        def store_permission(
            x: SyftVerifyKey | None = None,
        ) -> ActionObjectPermission:
            return ActionObjectPermission(result_id, read_permission, x)

        def blob_permission(
            x: SyftVerifyKey | None = None,
        ) -> ActionObjectPermission:
            return ActionObjectPermission(result_blob_id, read_permission, x)

        if len(output_readers) > 0:
            store_permissions = [store_permission(x) for x in output_readers]
            self.store.add_permissions(store_permissions)

            if result_blob_id is not None:
                blob_permissions = [blob_permission(x) for x in output_readers]
                blob_storage_service.stash.add_permissions(blob_permissions)

        return set_result

    def execute_plan(
        self,
        plan: Any,
        context: AuthedServiceContext,
        plan_kwargs: dict[str, ActionObject],
    ) -> Result[ActionObject, str] | SyftError:
        id2inpkey = {v.id: k for k, v in plan.inputs.items()}

        for plan_action in plan.actions:
            if (
                hasattr(plan_action.remote_self, "id")
                and plan_action.remote_self.id in id2inpkey
            ):
                plan_action.remote_self = plan_kwargs[
                    id2inpkey[plan_action.remote_self.id]
                ]
            for i, arg in enumerate(plan_action.args):
                if arg in id2inpkey:
                    plan_action.args[i] = plan_kwargs[id2inpkey[arg]]

            for k, arg in enumerate(plan_action.kwargs):
                if arg in id2inpkey:
                    plan_action.kwargs[k] = plan_kwargs[id2inpkey[arg]]

        for plan_action in plan.actions:
            action_res = self.execute(context, plan_action)
            if isinstance(action_res, SyftError):
                return action_res
        result_id = plan.outputs[0].id
        return self._get(context, result_id, TwinMode.MOCK, has_permission=True)

    def call_function(
        self, context: AuthedServiceContext, action: Action
    ) -> Result[ActionObject, str] | Err:
        # run function/class init
        _user_lib_config_registry = UserLibConfigRegistry.from_user(context.credentials)
        absolute_path = f"{action.path}.{action.op}"
        if absolute_path in _user_lib_config_registry:
            # TODO: implement properly
            # Now we are assuming its a function/class
            return execute_callable(self, context, action)
        else:
            return Err(
                f"Failed executing {action}. You have no permission for {absolute_path}"
            )

    def set_attribute(
        self,
        context: AuthedServiceContext,
        action: Action,
        resolved_self: ActionObject | TwinObject,
    ) -> Result[TwinObject | ActionObject, str]:
        args, _ = resolve_action_args(action, context, self)
        if args.is_err():
            return Err(
                f"Failed executing action {action}, could not resolve args: {args.err()}"
            )
        else:
            args = args.ok()
        if not isinstance(args[0], ActionObject):
            return Err(
                f"Failed executing action {action} setattribute requires a non-twin string as first argument"
            )
        name = args[0].syft_action_data
        # dont do the whole filtering dance with the name
        args = [args[1]]

        if isinstance(resolved_self, TwinObject):
            # todo, create copy?
            private_args = filter_twin_args(args, twin_mode=TwinMode.PRIVATE)
            private_val = private_args[0]
            setattr(resolved_self.private.syft_action_data, name, private_val)
            # todo: what do we use as data for the mock here?
            # depending on permisisons?
            public_args = filter_twin_args(args, twin_mode=TwinMode.MOCK)
            public_val = public_args[0]
            setattr(resolved_self.mock.syft_action_data, name, public_val)
            return Ok(
                TwinObject(
                    id=action.result_id,
                    private_obj=ActionObject.from_obj(
                        resolved_self.private.syft_action_data
                    ),
                    private_obj_id=action.result_id,
                    mock_obj=ActionObject.from_obj(resolved_self.mock.syft_action_data),
                    mock_obj_id=action.result_id,
                )
            )
        else:
            # TODO: Implement for twinobject args
            args = filter_twin_args(args, twin_mode=TwinMode.NONE)  # type: ignore[unreachable]
            val = args[0]
            setattr(resolved_self.syft_action_data, name, val)
            return Ok(
                ActionObject.from_obj(resolved_self.syft_action_data),
            )
            # todo: permissions
            # setattr(resolved_self.syft_action_data, name, val)
            # val = resolved_self.syft_action_data
            # result_action_object = Ok(wrap_result(action.result_id, val))

    def get_attribute(
        self, action: Action, resolved_self: ActionObject | TwinObject
    ) -> Ok[TwinObject | ActionObject]:
        if isinstance(resolved_self, TwinObject):
            private_result = getattr(resolved_self.private.syft_action_data, action.op)
            mock_result = getattr(resolved_self.mock.syft_action_data, action.op)
            return Ok(
                TwinObject(
                    id=action.result_id,
                    private_obj=ActionObject.from_obj(private_result),
                    private_obj_id=action.result_id,
                    mock_obj=ActionObject.from_obj(mock_result),
                    mock_obj_id=action.result_id,
                )
            )
        else:
            val = getattr(resolved_self.syft_action_data, action.op)  # type: ignore[unreachable]
            return Ok(wrap_result(action.result_id, val))

    def call_method(
        self,
        context: AuthedServiceContext,
        action: Action,
        resolved_self: ActionObject | TwinObject,
    ) -> Result[TwinObject | Any, str]:
        if isinstance(resolved_self, TwinObject):
            # method
            private_result = execute_object(
                self,
                context,
                resolved_self.private,
                action,
                twin_mode=TwinMode.PRIVATE,
            )
            if private_result.is_err():
                return Err(
                    f"Failed executing action {action}, result is an error: {private_result.err()}"
                )
            mock_result = execute_object(
                self, context, resolved_self.mock, action, twin_mode=TwinMode.MOCK
            )
            if mock_result.is_err():
                return Err(
                    f"Failed executing action {action}, result is an error: {mock_result.err()}"
                )

            private_result = private_result.ok()
            mock_result = mock_result.ok()

            return Ok(
                TwinObject(
                    id=action.result_id,
                    private_obj=private_result,
                    private_obj_id=action.result_id,
                    mock_obj=mock_result,
                    mock_obj_id=action.result_id,
                )
            )
        else:
            return execute_object(self, context, resolved_self, action)  # type:ignore[unreachable]

    def unwrap_nested_actionobjects(
        self, context: AuthedServiceContext, data: Any
    ) -> Any:
        """recursively unwraps nested action objects"""

        if isinstance(data, list):
            return [self.unwrap_nested_actionobjects(context, obj) for obj in data]
        if isinstance(data, dict):
            return {
                key: self.unwrap_nested_actionobjects(context, obj)
                for key, obj in data.items()
            }
        if isinstance(data, ActionObject):
            res = self.get(context=context, uid=data.id)
            res = res.ok() if res.is_ok() else res.err()
            if not isinstance(res, ActionObject):
                return SyftError(message=f"{res}")
            else:
                nested_res = res.syft_action_data
                if isinstance(nested_res, ActionObject):
                    raise ValueError(
                        "More than double nesting of ActionObjects is currently not supported"
                    )
                return nested_res
        return data

    def contains_nested_actionobjects(self, data: Any) -> bool:
        """
        returns if this is a list/set/dict that contains ActionObjects
        """

        def unwrap_collection(col: set | dict | list) -> [Any]:  # type: ignore
            return_values = []
            if isinstance(col, dict):
                values = list(col.values()) + list(col.keys())
            else:
                values = list(col)
            for v in values:
                if isinstance(v, list | dict | set):
                    return_values += unwrap_collection(v)
                else:
                    return_values.append(v)
            return return_values

        if isinstance(data, list | dict | set):
            values = unwrap_collection(data)
            has_action_object = any(isinstance(x, ActionObject) for x in values)
            return has_action_object
        elif isinstance(data, ActionObject):
            return True
        return False

    def flatten_action_arg(self, context: AuthedServiceContext, arg: UID) -> UID | None:
        """ "If the argument is a collection (of collections) of ActionObjects,
        We want to flatten the collection and upload a new ActionObject that contains
        its values. E.g. [[ActionObject1, ActionObject2],[ActionObject3, ActionObject4]]
        -> [[value1, value2],[value3, value4]]
        """
        res = self.get(context=context, uid=arg)
        if res.is_err():
            return arg

        action_object = res.ok()
        data = action_object.syft_action_data

        if self.contains_nested_actionobjects(data):
            new_data = self.unwrap_nested_actionobjects(context, data)
            # Update existing action object with the new flattened data
            action_object.syft_action_data_cache = new_data
            action_object._save_to_blob_storage()
            res = self._set(
                context=context,
                action_object=action_object,
            )

        return None

    @service_method(path="action.execute", name="execute", roles=GUEST_ROLE_LEVEL)
    def execute(
        self, context: AuthedServiceContext, action: Action
    ) -> Result[ActionObject, Err]:
        """Execute an operation on objects in the action store"""
        # relative
        from .plan import Plan

        if action.action_type == ActionType.CREATEOBJECT:
            result_action_object = Ok(action.create_object)
        elif action.action_type == ActionType.SYFTFUNCTION:
            usercode_service = context.server.get_service("usercodeservice")
            kwarg_ids = {}
            for k, v in action.kwargs.items():
                # transform lineage ids into ids
                kwarg_ids[k] = v.id
            result_action_object = usercode_service._call(
                context, action.user_code_id, action.result_id, **kwarg_ids
            )
            return result_action_object
        elif action.action_type == ActionType.FUNCTION:
            result_action_object = self.call_function(context, action)
        else:
            resolved_self = self._get(
                context=context,
                uid=action.remote_self,
                twin_mode=TwinMode.NONE,
                has_permission=True,
            )
            if resolved_self.is_err():
                return Err(
                    f"Failed executing action {action}, could not resolve self: {resolved_self.err()}"
                )
            resolved_self = resolved_self.ok()
            if action.op == "__call__" and resolved_self.syft_action_data_type == Plan:
                result_action_object = self.execute_plan(
                    plan=resolved_self.syft_action_data,
                    context=context,
                    plan_kwargs=action.kwargs,
                )
                return result_action_object
            elif action.action_type == ActionType.SETATTRIBUTE:
                result_action_object = self.set_attribute(
                    context, action, resolved_self
                )
            elif action.action_type == ActionType.GETATTRIBUTE:
                result_action_object = self.get_attribute(action, resolved_self)
            elif action.action_type == ActionType.METHOD:
                result_action_object = self.call_method(context, action, resolved_self)
            else:
                return Err("Unknown action")

        if result_action_object.is_err():
            return Err(
                f"Failed executing action {action}, result is an error: {result_action_object.err()}"
            )
        else:
            result_action_object = result_action_object.ok()

        # check if we have read permissions on the result
        has_result_read_permission = self.has_read_permission_for_action_result(
            context, action
        )

        result_action_object._set_obj_location_(
            context.server.id,
            context.credentials,
        )
        blob_store_result = result_action_object._save_to_blob_storage()
        if isinstance(blob_store_result, SyftError):
            return blob_store_result

        # pass permission information to the action store as extra kwargs
        context.extra_kwargs = {
            "has_result_read_permission": has_result_read_permission
        }
        if isinstance(blob_store_result, SyftWarning):
            logger.debug(blob_store_result.message)
        set_result = self._set(
            context,
            result_action_object,
        )
        if set_result.is_err():
            return Err(
                f"Failed executing action {action}, set result is an error: {set_result.err()}"
            )

        return set_result

    def has_read_permission_for_action_result(
        self, context: AuthedServiceContext, action: Action
    ) -> bool:
        action_obj_ids = (
            action.args + list(action.kwargs.values()) + [action.remote_self]
        )
        permissions = [
            ActionObjectREAD(uid=_id, credentials=context.credentials)
            for _id in action_obj_ids
        ]
        return self.store.has_permissions(permissions)

    @service_method(path="action.exists", name="exists", roles=GUEST_ROLE_LEVEL)
    def exists(
        self, context: AuthedServiceContext, obj_id: UID
    ) -> Result[SyftSuccess, SyftError]:
        """Checks if the given object id exists in the Action Store"""
        if self.store.exists(obj_id):
            return SyftSuccess(message=f"Object: {obj_id} exists")
        else:
            return SyftError(message=f"Object: {obj_id} does not exist")

    @service_method(path="action.delete", name="delete", roles=ADMIN_ROLE_LEVEL)
    def delete(
        self, context: AuthedServiceContext, uid: UID, soft_delete: bool = False
    ) -> SyftSuccess | SyftError:
        get_res = self.store.get(uid=uid, credentials=context.credentials)
        if get_res.is_err():
            return SyftError(message=get_res.err())
        obj: ActionObject | TwinObject = get_res.ok()
        return_msg = []

        # delete any associated blob storage entry object to the action object
        blob_del_res = self._delete_blob_storage_entry(context=context, obj=obj)
        if isinstance(blob_del_res, SyftError):
            return SyftError(message=blob_del_res.message)
        return_msg.append(blob_del_res.message)

        # delete the action object from the action store
        store_del_res = self._delete_from_action_store(
            context=context, uid=obj.id, soft_delete=soft_delete
        )
        if isinstance(store_del_res, SyftError):
            return SyftError(message=store_del_res.message)
        return_msg.append(store_del_res.message)

        return SyftSuccess(message="\n".join(return_msg))

    def _delete_blob_storage_entry(
        self,
        context: AuthedServiceContext,
        obj: TwinObject | ActionObject,
    ) -> SyftSuccess | SyftError:
        deleted_blob_ids = []
        blob_store_service = cast(
            BlobStorageService, context.server.get_service(BlobStorageService)
        )

        if isinstance(obj, ActionObject) and obj.syft_blob_storage_entry_id:
            blob_del_res = blob_store_service.delete(
                context=context, uid=obj.syft_blob_storage_entry_id
            )
            if isinstance(blob_del_res, SyftError):
                return SyftError(message=blob_del_res.message)
            deleted_blob_ids.append(obj.syft_blob_storage_entry_id)

        if isinstance(obj, TwinObject):
            if obj.private.syft_blob_storage_entry_id:
                blob_del_res = blob_store_service.delete(
                    context=context, uid=obj.private.syft_blob_storage_entry_id
                )
                if isinstance(blob_del_res, SyftError):
                    return SyftError(message=blob_del_res.message)
                deleted_blob_ids.append(obj.private.syft_blob_storage_entry_id)

            if obj.mock.syft_blob_storage_entry_id:
                blob_del_res = blob_store_service.delete(
                    context=context, uid=obj.mock.syft_blob_storage_entry_id
                )
                if isinstance(blob_del_res, SyftError):
                    return SyftError(message=blob_del_res.message)
                deleted_blob_ids.append(obj.mock.syft_blob_storage_entry_id)

        message = f"Deleted blob storage entries: {', '.join(str(blob_id) for blob_id in deleted_blob_ids)}"

        return SyftSuccess(message=message)

    def _delete_from_action_store(
        self,
        context: AuthedServiceContext,
        uid: UID,
        soft_delete: bool = False,
    ) -> SyftSuccess | SyftError:
        if soft_delete:
            get_res = self.store.get(uid=uid, credentials=context.credentials)
            if get_res.is_err():
                return SyftError(message=get_res.err())
            obj: ActionObject | TwinObject = get_res.ok()

            if isinstance(obj, TwinObject):
                res = self._soft_delete_action_obj(
                    context=context, action_obj=obj.private
                )
                if res.is_err():
                    return SyftError(message=res.err())
                res = self._soft_delete_action_obj(context=context, action_obj=obj.mock)
                if res.is_err():
                    return SyftError(message=res.err())

            if isinstance(obj, ActionObject):
                res = self._soft_delete_action_obj(context=context, action_obj=obj)
                if res.is_err():
                    return SyftError(message=res.err())
        else:
            res = self.store.delete(credentials=context.credentials, uid=uid)
            if res.is_err():
                return SyftError(message=res.err())

        return SyftSuccess(message=f"Action object with uid '{uid}' deleted.")

    def _soft_delete_action_obj(
        self, context: AuthedServiceContext, action_obj: ActionObject
    ) -> Result[ActionObject, str]:
        action_obj.syft_action_data_cache = None
        res = action_obj._save_to_blob_storage()
        if isinstance(res, SyftError):
            return Err(res.message)
        set_result = self._set(
            context=context,
            action_object=action_obj,
        )
        return set_result


def resolve_action_args(
    action: Action, context: AuthedServiceContext, service: ActionService
) -> tuple[Ok[dict], bool]:
    has_twin_inputs = False
    args = []
    for arg_id in action.args:
        arg_value = service._get(
            context=context, uid=arg_id, twin_mode=TwinMode.NONE, has_permission=True
        )
        if arg_value.is_err():
            return arg_value, False
        if isinstance(arg_value.ok(), TwinObject):
            has_twin_inputs = True
        args.append(arg_value.ok())
    return Ok(args), has_twin_inputs


def resolve_action_kwargs(
    action: Action, context: AuthedServiceContext, service: ActionService
) -> tuple[Ok[dict], bool]:
    has_twin_inputs = False
    kwargs = {}
    for key, arg_id in action.kwargs.items():
        kwarg_value = service._get(
            context=context, uid=arg_id, twin_mode=TwinMode.NONE, has_permission=True
        )
        if kwarg_value.is_err():
            return kwarg_value, False
        if isinstance(kwarg_value.ok(), TwinObject):
            has_twin_inputs = True
        kwargs[key] = kwarg_value.ok()
    return Ok(kwargs), has_twin_inputs


def execute_callable(
    service: ActionService,
    context: AuthedServiceContext,
    action: Action,
) -> Result[ActionObject, str]:
    args, has_arg_twins = resolve_action_args(action, context, service)
    kwargs, has_kwargs_twins = resolve_action_kwargs(action, context, service)
    has_twin_inputs = has_arg_twins or has_kwargs_twins
    if args.is_err():
        return args
    else:
        args = args.ok()
    if kwargs.is_err():
        return kwargs
    else:
        kwargs = kwargs.ok()

    # 🔵 TODO 10: Get proper code From old RunClassMethodAction to ensure the function
    # is not bound to the original object or mutated
    # stdlib

    # TODO: get from CMPTree is probably safer
    def _get_target_callable(path: str, op: str) -> Any:
        path_elements = path.split(".")
        res = importlib.import_module(path_elements[0])
        for p in path_elements[1:]:
            res = getattr(res, p)
        res = getattr(res, op)
        return res

    target_callable = _get_target_callable(action.path, action.op)

    result = None
    try:
        if target_callable:
            if not has_twin_inputs:
                # if twin_mode == TwinMode.NONE and not has_twin_inputs:
                twin_mode = TwinMode.NONE
                # no twins
                filtered_args = filter_twin_args(args, twin_mode=twin_mode)
                filtered_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                result = target_callable(*filtered_args, **filtered_kwargs)
                result_action_object = wrap_result(action.result_id, result)
            else:
                twin_mode = TwinMode.PRIVATE
                private_args = filter_twin_args(args, twin_mode=twin_mode)
                private_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                private_result = target_callable(*private_args, **private_kwargs)
                result_action_object_private = wrap_result(
                    action.result_id, private_result
                )

                twin_mode = TwinMode.MOCK
                mock_args = filter_twin_args(args, twin_mode=twin_mode)
                mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                mock_result = target_callable(*mock_args, **mock_kwargs)
                result_action_object_mock = wrap_result(action.result_id, mock_result)

                result_action_object = TwinObject(
                    id=action.result_id,
                    private_obj=result_action_object_private,
                    mock_obj=result_action_object_mock,
                )

    except Exception as e:
        print("what is this exception", e)
        return Err(e)
    return Ok(result_action_object)


def execute_object(
    service: ActionService,
    context: AuthedServiceContext,
    resolved_self: ActionObject,
    action: Action,
    twin_mode: TwinMode = TwinMode.NONE,
) -> Result[Ok[TwinObject | ActionObject], Err[str]]:
    unboxed_resolved_self = resolved_self.syft_action_data
    _args, has_arg_twins = resolve_action_args(action, context, service)

    kwargs, has_kwargs_twins = resolve_action_kwargs(action, context, service)
    if _args.is_err():
        return _args
    else:
        args = _args.ok()
    if kwargs.is_err():
        return kwargs
    else:
        kwargs = kwargs.ok()
    has_twin_inputs = has_arg_twins or has_kwargs_twins

    # 🔵 TODO 10: Get proper code From old RunClassMethodAction to ensure the function
    # is not bound to the original object or mutated
    target_method = getattr(unboxed_resolved_self, action.op, None)
    result = None
    try:
        if target_method:
            if twin_mode == TwinMode.NONE and not has_twin_inputs:
                # no twins
                filtered_args = filter_twin_args(args, twin_mode=twin_mode)
                filtered_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                result = target_method(*filtered_args, **filtered_kwargs)
                result_action_object = wrap_result(action.result_id, result)
            elif twin_mode == TwinMode.NONE and has_twin_inputs:
                # self isn't a twin but one of the inputs is
                private_args = filter_twin_args(args, twin_mode=TwinMode.PRIVATE)
                private_kwargs = filter_twin_kwargs(kwargs, twin_mode=TwinMode.PRIVATE)
                private_result = target_method(*private_args, **private_kwargs)
                result_action_object_private = wrap_result(
                    action.result_id, private_result
                )

                mock_args = filter_twin_args(args, twin_mode=TwinMode.MOCK)
                mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=TwinMode.MOCK)
                mock_result = target_method(*mock_args, **mock_kwargs)
                result_action_object_mock = wrap_result(action.result_id, mock_result)

                result_action_object = TwinObject(
                    id=action.result_id,
                    private_obj=result_action_object_private,
                    mock_obj=result_action_object_mock,
                )
            elif twin_mode == twin_mode.PRIVATE:  # type:ignore
                # twin private path
                private_args = filter_twin_args(args, twin_mode=twin_mode)  # type:ignore[unreachable]
                private_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                result = target_method(*private_args, **private_kwargs)
                result_action_object = wrap_result(action.result_id, result)
            elif twin_mode == twin_mode.MOCK:  # type:ignore
                # twin mock path
                mock_args = filter_twin_args(args, twin_mode=twin_mode)  # type:ignore[unreachable]
                mock_kwargs = filter_twin_kwargs(kwargs, twin_mode=twin_mode)
                target_method = getattr(unboxed_resolved_self, action.op, None)
                result = target_method(*mock_args, **mock_kwargs)
                result_action_object = wrap_result(action.result_id, result)
            else:
                raise Exception(
                    f"Bad combination of: twin_mode: {twin_mode} and has_twin_inputs: {has_twin_inputs}"
                )
        else:
            return Err("Missing target method")

    except Exception as e:
        return Err(e)

    return Ok(result_action_object)


def wrap_result(result_id: UID, result: Any) -> ActionObject:
    # 🟡 TODO 11: Figure out how we want to store action object results
    action_type = action_type_for_type(result)
    result_action_object = action_type(id=result_id, syft_action_data_cache=result)
    return result_action_object


def filter_twin_args(args: list[Any], twin_mode: TwinMode) -> Any:
    filtered = []
    for arg in args:
        if isinstance(arg, TwinObject):
            if twin_mode == TwinMode.PRIVATE:
                filtered.append(arg.private.syft_action_data)
            elif twin_mode == TwinMode.MOCK:
                filtered.append(arg.mock.syft_action_data)
            else:
                raise Exception(
                    f"Filter can only use {TwinMode.PRIVATE} or {TwinMode.MOCK}"
                )
        else:
            filtered.append(arg.syft_action_data)
    return filtered


def filter_twin_kwargs(
    kwargs: dict, twin_mode: TwinMode, allow_python_types: bool = False
) -> Any:
    filtered = {}
    for k, v in kwargs.items():
        if isinstance(v, TwinObject):
            if twin_mode == TwinMode.PRIVATE:
                filtered[k] = v.private.syft_action_data
            elif twin_mode == TwinMode.MOCK:
                filtered[k] = v.mock.syft_action_data
            else:
                raise Exception(
                    f"Filter can only use {TwinMode.PRIVATE} or {TwinMode.MOCK}"
                )
        else:
            if isinstance(v, ActionObject):
                filtered[k] = v.syft_action_data
            elif (
                isinstance(v, str | int | float | dict | CustomEndpointActionObject)
                and allow_python_types
            ):
                filtered[k] = v
            else:
                # third party
                raise ValueError(
                    f"unexepected value {v} passed to filtered twin kwargs"
                )
    return filtered


TYPE_TO_SERVICE[ActionObject] = ActionService
TYPE_TO_SERVICE[TwinObject] = ActionService
TYPE_TO_SERVICE[AnyActionObject] = ActionService

SERVICE_TO_TYPES[ActionService].update({ActionObject, TwinObject, AnyActionObject})



# stdlib

# stdlib
from typing import TypeVar

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ..context import AuthedServiceContext
from ..notification.notifications import Notification
from ..response import SyftError
from ..response import SyftSuccess
from .notifier_enums import NOTIFIERS
from .smtp_client import SMTPClient


class BaseNotifier:
    def send(
        self, target: SyftVerifyKey, notification: Notification
    ) -> SyftSuccess | SyftError:
        return SyftError(message="Not implemented")


TBaseNotifier = TypeVar("TBaseNotifier", bound=BaseNotifier)


@serializable(canonical_name="EmailNotifier", version=1)
class EmailNotifier(BaseNotifier):
    smtp_client: SMTPClient
    sender = ""

    def __init__(
        self,
        username: str,
        password: str,
        sender: str,
        server: str,
        port: int = 587,
    ) -> None:
        self.sender = sender
        self.smtp_client = SMTPClient(
            server=server,
            port=port,
            username=username,
            password=password,
        )

    @classmethod
    def check_credentials(
        cls,
        username: str,
        password: str,
        server: str,
        port: int = 587,
    ) -> Result[Ok, Err]:
        return SMTPClient.check_credentials(
            server=server,
            port=port,
            username=username,
            password=password,
        )

    def send(
        self, context: AuthedServiceContext, notification: Notification
    ) -> Result[Ok, Err]:
        try:
            user_service = context.server.get_service("userservice")

            receiver = user_service.get_by_verify_key(notification.to_user_verify_key)

            if not receiver.notifications_enabled[NOTIFIERS.EMAIL]:
                return Ok(
                    "Email notifications are disabled for this user."
                )  # TODO: Should we return an error here?

            receiver_email = receiver.email

            if notification.email_template:
                subject = notification.email_template.email_title(
                    notification, context=context
                )
                body = notification.email_template.email_body(
                    notification, context=context
                )
            else:
                subject = notification.subject
                body = notification._repr_html_()

            if isinstance(receiver_email, str):
                receiver_email = [receiver_email]

            self.smtp_client.send(
                sender=self.sender, receiver=receiver_email, subject=subject, body=body
            )
            return Ok("Email sent successfully!")
        except Exception:
            return Err(
                "Some notifications failed to be delivered. Please check the health of the mailing server."
            )


@serializable()
class NotificationPreferences(SyftObject):
    __canonical_name__ = "NotificationPreferences"
    __version__ = SYFT_OBJECT_VERSION_1
    __repr_attrs__ = [
        "email",
        "sms",
        "slack",
        "app",
    ]

    email: bool = False
    sms: bool = False
    slack: bool = False
    app: bool = False


@serializable()
class NotifierSettings(SyftObject):
    __canonical_name__ = "NotifierSettings"
    __version__ = SYFT_OBJECT_VERSION_1
    __repr_attrs__ = [
        "active",
        "email_enabled",
    ]
    active: bool = False
    # Flag to identify which notification is enabled
    # For now, consider only the email notification
    # In future, Admin, must be able to have a better
    # control on diff notifications.

    notifiers: dict[NOTIFIERS, type[TBaseNotifier]] = {
        NOTIFIERS.EMAIL: EmailNotifier,
    }

    notifiers_status: dict[NOTIFIERS, bool] = {
        NOTIFIERS.EMAIL: True,
        NOTIFIERS.SMS: False,
        NOTIFIERS.SLACK: False,
        NOTIFIERS.APP: False,
    }

    email_sender: str | None = ""
    email_server: str | None = ""
    email_port: int | None = 587
    email_username: str | None = ""
    email_password: str | None = ""

    @property
    def email_enabled(self) -> bool:
        return self.notifiers_status[NOTIFIERS.EMAIL]

    @property
    def sms_enabled(self) -> bool:
        return self.notifiers_status[NOTIFIERS.SMS]

    @property
    def slack_enabled(self) -> bool:
        return self.notifiers_status[NOTIFIERS.SLACK]

    @property
    def app_enabled(self) -> bool:
        return self.notifiers_status[NOTIFIERS.APP]

    def validate_email_credentials(
        self,
        username: str,
        password: str,
        server: str,
        port: int,
    ) -> Result[Ok, Err]:
        return self.notifiers[NOTIFIERS.EMAIL].check_credentials(
            server=server,
            port=port,
            username=username,
            password=password,
        )

    def send_notifications(
        self,
        context: AuthedServiceContext,
        notification: Notification,
    ) -> Result[Ok, Err]:
        notifier_objs: list = self.select_notifiers(notification)

        for notifier in notifier_objs:
            result = notifier.send(context, notification)
            if result.err():
                return result

        return Ok("Notification sent successfully!")

    def select_notifiers(self, notification: Notification) -> list[BaseNotifier]:
        """
        Return a list of the notifiers enabled for the given notification"

        Args:
            notification (Notification): The notification object
        Returns:
            List[BaseNotifier]: A list of enabled notifier objects
        """
        notifier_objs = []
        for notifier_type in notification.notifier_types:
            # Check if the notifier is enabled and if it is, create the notifier object
            if (
                self.notifiers_status[notifier_type]
                and self.notifiers[notifier_type] is not None
            ):
                # If notifier is email, we need to pass the parameters
                if notifier_type == NOTIFIERS.EMAIL:
                    notifier_objs.append(
                        self.notifiers[notifier_type](  # type: ignore[misc]
                            username=self.email_username,
                            password=self.email_password,
                            sender=self.email_sender,
                            server=self.email_server,
                        )
                    )
                # If notifier is not email, we just create the notifier object
                # TODO: Add the other notifiers, and its auth methods
                else:
                    notifier_objs.append(self.notifiers[notifier_type]())  # type: ignore[misc]

        return notifier_objs



# stdlib
import secrets
from typing import cast

# third party
from result import Err
from result import Ok
from result import Result

# relative
from ...client.client import SyftClient
from ...serde.serializable import serializable
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ..context import ChangeContext
from ..request.request import Change
from ..response import SyftError
from ..response import SyftSuccess
from .routes import ServerRoute
from .server_peer import ServerPeer


@serializable()
class AssociationRequestChange(Change):
    __canonical_name__ = "AssociationRequestChange"
    __version__ = SYFT_OBJECT_VERSION_1

    self_server_route: ServerRoute
    remote_peer: ServerPeer
    challenge: bytes

    __repr_attrs__ = ["self_server_route", "remote_peer"]

    def _run(
        self, context: ChangeContext, apply: bool
    ) -> Result[tuple[bytes, ServerPeer], SyftError]:
        """
        Executes the association request.

        Args:
            context (ChangeContext): The change context.
            apply (bool): A flag indicating whether to apply the association request.

        Returns:
            Result[tuple[bytes, ServerPeer], SyftError]: The result of the association request.
        """
        # relative
        from .network_service import NetworkService

        if not apply:
            # TODO: implement undo for AssociationRequestChange
            return Err(
                SyftError(message="Undo not supported for AssociationRequestChange")
            )

        # Get the network service
        service_ctx = context.to_service_ctx()
        network_service = cast(
            NetworkService, service_ctx.server.get_service(NetworkService)
        )
        network_stash = network_service.stash

        # Check if remote peer to be added is via reverse tunnel
        rtunnel_route = self.remote_peer.get_rtunnel_route()
        add_rtunnel_route = (
            rtunnel_route is not None
            and self.remote_peer.latest_added_route == rtunnel_route
        )

        # If the remote peer is added via reverse tunnel, we skip ping to peer
        if add_rtunnel_route:
            network_service.set_reverse_tunnel_config(
                context=context,
                remote_server_peer=self.remote_peer,
            )
        else:
            # Pinging the remote peer to verify the connection
            try:
                remote_client: SyftClient = self.remote_peer.client_with_context(
                    context=service_ctx
                )
                if remote_client.is_err():
                    return SyftError(
                        message=f"Failed to create remote client for peer: "
                        f"{self.remote_peer.id}. Error: {remote_client.err()}"
                    )
                remote_client = remote_client.ok()
                random_challenge = secrets.token_bytes(16)
                remote_res = remote_client.api.services.network.ping(
                    challenge=random_challenge
                )
            except Exception as e:
                return SyftError(message="Remote Peer cannot ping peer:" + str(e))

            if isinstance(remote_res, SyftError):
                return Err(remote_res)

            challenge_signature = remote_res

            # Verifying if the challenge is valid
            try:
                self.remote_peer.verify_key.verify_key.verify(
                    random_challenge, challenge_signature
                )
            except Exception as e:
                return Err(SyftError(message=str(e)))

        # Adding the remote peer to the network stash
        result = network_stash.create_or_update_peer(
            service_ctx.server.verify_key, self.remote_peer
        )

        if result.is_err():
            return Err(SyftError(message=str(result.err())))

        # this way they can match up who we are with who they think we are
        # Sending a signed messages for the peer to verify
        self_server_peer = self.self_server_route.validate_with_context(
            context=service_ctx
        )

        if isinstance(self_server_peer, SyftError):
            return Err(self_server_peer)

        return Ok(
            SyftSuccess(
                message=f"Routes successfully added for peer: {self.remote_peer.name}"
            )
        )

    def apply(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context, apply=True)

    def undo(self, context: ChangeContext) -> Result[SyftSuccess, SyftError]:
        return self._run(context, apply=False)

    def __repr_syft_nested__(self) -> str:
        return f"Request for connection from : {self.remote_peer.name}"



# stdlib
from typing import Any
from typing import ClassVar

# relative
from ...serde.serializable import serializable
from ...service.context import AuthedServiceContext
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syncable_object import SyncableSyftObject
from ...types.uid import UID


@serializable()
class SyftLog(SyncableSyftObject):
    __canonical_name__ = "SyftLog"
    __version__ = SYFT_OBJECT_VERSION_1

    __repr_attrs__ = ["stdout", "stderr"]
    __exclude_sync_diff_attrs__: list[str] = []
    __private_sync_attr_mocks__: ClassVar[dict[str, Any]] = {
        "stderr": "",
        "stdout": "",
    }

    stdout: str = ""
    stderr: str = ""
    job_id: UID

    def append(self, new_str: str) -> None:
        self.stdout += new_str

    def append_error(self, new_str: str) -> None:
        self.stderr += new_str

    def restart(self) -> None:
        self.stderr = ""
        self.stdout = ""

    def get_sync_dependencies(
        self, context: AuthedServiceContext, **kwargs: dict
    ) -> list[UID]:  # type: ignore
        return [self.job_id]



# stdlib
from typing import Any

# third party
from capnp.lib.capnp import _DynamicStructBuilder


def _deserialize(
    blob: Any,
    from_proto: bool = True,
    from_bytes: bool = False,
) -> Any:
    # relative
    from .recursive import rs_bytes2object
    from .recursive import rs_proto2object

    if (
        (from_bytes and not isinstance(blob, bytes))
        or (
            from_proto
            and not from_bytes
            and not isinstance(blob, _DynamicStructBuilder)
        )
        or not (from_bytes or from_proto)
    ):
        raise TypeError("Wrong deserialization format.")

    if from_bytes:
        return rs_bytes2object(blob)

    if from_proto:
        return rs_proto2object(blob)



# stdlib
import logging
import threading
import time
from typing import cast

# relative
from ...serde.serializable import serializable
from ...types.datetime import DateTime
from ..context import AuthedServiceContext
from ..response import SyftError
from .network_service import NetworkService
from .network_service import ServerPeerAssociationStatus
from .server_peer import ServerPeer
from .server_peer import ServerPeerConnectionStatus
from .server_peer import ServerPeerUpdate

logger = logging.getLogger(__name__)


@serializable(without=["thread"], canonical_name="PeerHealthCheckTask", version=1)
class PeerHealthCheckTask:
    repeat_time = 10  # in seconds

    def __init__(self) -> None:
        self.thread: threading.Thread | None = None
        self.started_time = None
        self._stop = False

    def peer_route_heathcheck(self, context: AuthedServiceContext) -> SyftError | None:
        """
        Perform a health check on the peers in the network stash.
        - If peer is accessible, ping the peer.
        - Peer is connected to the network.

        Args:
            context (AuthedServiceContext): The authenticated service context.

        Returns:
            None
        """

        network_service = cast(
            NetworkService, context.server.get_service(NetworkService)
        )
        network_stash = network_service.stash

        result = network_stash.get_all(context.server.verify_key)

        if result.is_err():
            logger.error(f"Failed to fetch peers from stash: {result.err()}")
            return SyftError(message=f"{result.err()}")

        all_peers: list[ServerPeer] = result.ok()

        for peer in all_peers:
            peer_update = ServerPeerUpdate(id=peer.id)
            peer_update.pinged_timestamp = DateTime.now()
            try:
                peer_client = peer.client_with_context(context=context)
                if peer_client.is_err():
                    logger.error(
                        f"Failed to create client for peer: {peer}: {peer_client.err()}"
                    )
                    peer_update.ping_status = ServerPeerConnectionStatus.TIMEOUT
                    peer_client = None
            except Exception as e:
                logger.error(f"Failed to create client for peer: {peer}", exc_info=e)

                peer_update.ping_status = ServerPeerConnectionStatus.TIMEOUT
                peer_client = None

            if peer_client is not None:
                peer_client = peer_client.ok()
                peer_status = peer_client.api.services.network.check_peer_association(
                    peer_id=context.server.id
                )
                peer_update.ping_status = (
                    ServerPeerConnectionStatus.ACTIVE
                    if peer_status == ServerPeerAssociationStatus.PEER_ASSOCIATED
                    else ServerPeerConnectionStatus.INACTIVE
                )
                if isinstance(peer_status, SyftError):
                    peer_update.ping_status_message = (
                        f"Error `{peer_status.message}` when pinging peer '{peer.name}'"
                    )
                else:
                    peer_update.ping_status_message = (
                        f"Peer '{peer.name}''s ping status: "
                        f"{peer_update.ping_status.value.lower()}"
                    )

            result = network_stash.update(
                credentials=context.server.verify_key,
                peer_update=peer_update,
                has_permission=True,
            )

            if result.is_err():
                logger.error(f"Failed to update peer in stash: {result.err()}")

        return None

    def _run(self, context: AuthedServiceContext) -> None:
        self.started_time = DateTime.now()
        while True:
            if self._stop:
                break
            self.peer_route_heathcheck(context)
            time.sleep(self.repeat_time)

    def run(self, context: AuthedServiceContext) -> None:
        if self.thread is not None:
            logger.info(
                f"Peer health check task is already running in thread "
                f"{self.thread.name} with ID: {self.thread.ident}."
            )
        else:
            self.thread = threading.Thread(target=self._run, args=(context,))
            logger.info(
                f"Start running peers health check in thread "
                f"{self.thread.name} with ID: {self.thread.ident}."
            )
            self.thread.start()

    def stop(self) -> None:
        if self.thread:
            self._stop = True
            self.thread.join()
            self.thread = None
            self.started_time = None
        logger.info("Peer health check task stopped.")



# third party
import ipywidgets as widgets

# relative
from ....types.syft_object import SYFT_OBJECT_VERSION_1
from ....types.syft_object import SyftBaseObject


class HTMLComponentBase(SyftBaseObject):
    __canonical_name__ = "HTMLComponentBase"
    __version__ = SYFT_OBJECT_VERSION_1

    def to_html(self) -> str:
        raise NotImplementedError()

    def to_widget(self) -> widgets.Widget:
        return widgets.HTML(value=self.to_html())

    def _repr_html_(self) -> str:
        return self.to_html()



# future
from __future__ import annotations

# stdlib
from collections.abc import Callable

# third party
from packaging import version
from pydantic import BaseModel
from pydantic import model_validator

# relative
from ...abstract_server import ServerType
from ...protocol.data_protocol import get_data_protocol
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import StorableObjectType
from ...types.syft_object import SyftObject
from ...types.transforms import convert_types
from ...types.transforms import drop
from ...types.transforms import rename
from ...types.transforms import transform
from ...types.uid import UID


def check_version(
    client_version: str, server_version: str, server_name: str, silent: bool = False
) -> bool:
    client_syft_version = version.parse(client_version)
    server_syft_version = version.parse(server_version)
    msg = (
        f"You are running syft=={client_version} but "
        f"{server_name} server requires {server_version}"
    )
    if client_syft_version.base_version != server_syft_version.base_version:
        raise ValueError(msg)
    if client_syft_version.pre != server_syft_version.pre:
        if not silent:
            print(f"Warning: {msg}")
            return False
    return True


@serializable()
class ServerMetadata(SyftObject):
    __canonical_name__ = "ServerMetadata"
    __version__ = SYFT_OBJECT_VERSION_1

    name: str
    id: UID
    verify_key: SyftVerifyKey
    highest_version: int
    lowest_version: int
    syft_version: str
    server_type: ServerType = ServerType.DATASITE
    organization: str = "OpenMined"
    description: str = "Text"
    server_side_type: str
    show_warnings: bool
    eager_execution_enabled: bool
    min_size_blob_storage_mb: int

    def check_version(self, client_version: str) -> bool:
        return check_version(
            client_version=client_version,
            server_version=self.syft_version,
            server_name=self.name,
        )


@serializable(canonical_name="ServerMetadataJSON", version=1)
class ServerMetadataJSON(BaseModel, StorableObjectType):
    metadata_version: int
    name: str
    id: str
    verify_key: str
    highest_object_version: int | None = None
    lowest_object_version: int | None = None
    syft_version: str
    server_type: str = ServerType.DATASITE.value
    organization: str = "OpenMined"
    description: str = "My cool datasite"
    signup_enabled: bool = False
    eager_execution_enabled: bool = False
    admin_email: str = ""
    server_side_type: str
    show_warnings: bool
    supported_protocols: list = []
    min_size_blob_storage_mb: int

    @model_validator(mode="before")
    @classmethod
    def add_protocol_versions(cls, values: dict) -> dict:
        if "supported_protocols" not in values:
            data_protocol = get_data_protocol()
            values["supported_protocols"] = data_protocol.supported_protocols
        return values

    def check_version(self, client_version: str) -> bool:
        return check_version(
            client_version=client_version,
            server_version=self.syft_version,
            server_name=self.name,
        )


@transform(ServerMetadata, ServerMetadataJSON)
def metadata_to_json() -> list[Callable]:
    return [
        drop(["__canonical_name__"]),
        rename("__version__", "metadata_version"),
        convert_types(["id", "verify_key", "server_type"], str),
        rename("highest_version", "highest_object_version"),
        rename("lowest_version", "lowest_object_version"),
    ]


@transform(ServerMetadataJSON, ServerMetadata)
def json_to_metadata() -> list[Callable]:
    return [
        drop(["metadata_version", "supported_protocols"]),
        convert_types(["id", "verify_key"], [UID, SyftVerifyKey]),
        convert_types(["server_type"], ServerType),
        rename("highest_object_version", "highest_version"),
        rename("lowest_object_version", "lowest_version"),
    ]



# stdlib
from typing import ClassVar

# third party
from pydantic import model_validator
from result import Err
from result import Ok
from result import Result

# relative
from ...client.api import APIRegistry
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...store.linked_obj import LinkedObject
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syncable_object import SyncableSyftObject
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_object import ActionObject
from ..action.action_permissions import ActionObjectREAD
from ..context import AuthedServiceContext
from ..response import SyftError
from ..service import AbstractService
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL

CreatedAtPartitionKey = PartitionKey(key="created_at", type_=DateTime)
UserCodeIdPartitionKey = PartitionKey(key="user_code_id", type_=UID)
JobIdPartitionKey = PartitionKey(key="job_id", type_=UID)
OutputPolicyIdPartitionKey = PartitionKey(key="output_policy_id", type_=UID)


@serializable()
class ExecutionOutput(SyncableSyftObject):
    __canonical_name__ = "ExecutionOutput"
    __version__ = SYFT_OBJECT_VERSION_1

    executing_user_verify_key: SyftVerifyKey
    user_code_link: LinkedObject
    output_ids: list[UID] | dict[str, UID] | None = None
    job_link: LinkedObject | None = None
    created_at: DateTime = DateTime.now()
    input_ids: dict[str, UID] | None = None

    # Required for __attr_searchable__, set by model_validator
    user_code_id: UID
    job_id: UID | None = None

    # Output policy is not a linked object because its saved on the usercode
    output_policy_id: UID | None = None

    __attr_searchable__: ClassVar[list[str]] = [
        "user_code_id",
        "created_at",
        "output_policy_id",
        "job_id",
    ]
    __repr_attrs__: ClassVar[list[str]] = [
        "created_at",
        "user_code_id",
        "job_id",
        "output_ids",
    ]

    @model_validator(mode="before")
    @classmethod
    def add_searchable_link_ids(cls, values: dict) -> dict:
        if "user_code_link" in values:
            values["user_code_id"] = values["user_code_link"].object_uid
        if values.get("job_link"):
            values["job_id"] = values["job_link"].object_uid
        return values

    @classmethod
    def from_ids(
        cls: type["ExecutionOutput"],
        output_ids: UID | list[UID] | dict[str, UID],
        user_code_id: UID,
        executing_user_verify_key: SyftVerifyKey,
        server_uid: UID,
        job_id: UID | None = None,
        output_policy_id: UID | None = None,
        input_ids: dict[str, UID] | None = None,
    ) -> "ExecutionOutput":
        # relative
        from ..code.user_code_service import UserCode
        from ..code.user_code_service import UserCodeService
        from ..job.job_service import Job
        from ..job.job_service import JobService

        if isinstance(output_ids, UID):
            output_ids = [output_ids]

        user_code_link = LinkedObject.from_uid(
            object_uid=user_code_id,
            object_type=UserCode,
            service_type=UserCodeService,
            server_uid=server_uid,
        )

        if job_id:
            job_link = LinkedObject.from_uid(
                object_uid=job_id,
                object_type=Job,
                service_type=JobService,
                server_uid=server_uid,
            )
        else:
            job_link = None
        return cls(
            output_ids=output_ids,
            user_code_link=user_code_link,
            job_link=job_link,
            executing_user_verify_key=executing_user_verify_key,
            output_policy_id=output_policy_id,
            input_ids=input_ids,
        )

    @property
    def outputs(self) -> list[ActionObject] | dict[str, ActionObject] | None:
        api = APIRegistry.api_for(
            server_uid=self.syft_server_location,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            raise ValueError(
                f"Can't access the api. Please log in to {self.syft_server_location}"
            )
        action_service = api.services.action

        # TODO: error handling for action_service.get
        if isinstance(self.output_ids, dict):
            return {k: action_service.get(v) for k, v in self.output_ids.items()}
        elif isinstance(self.output_ids, list):
            return [action_service.get(v) for v in self.output_ids]
        else:
            return None

    @property
    def output_id_list(self) -> list[UID]:
        ids = self.output_ids
        if isinstance(ids, dict):
            return list(ids.values())
        elif isinstance(ids, list):
            return ids
        return []

    @property
    def input_id_list(self) -> list[UID]:
        ids = self.input_ids
        if isinstance(ids, dict):
            return list(ids.values())
        return []

    def check_input_ids(self, kwargs: dict[str, UID]) -> bool:
        """
        Checks the input IDs against the stored input IDs.

        Args:
            kwargs (dict[str, UID]): A dictionary containing the input IDs to be checked.

        Returns:
            bool: True if the input IDs are valid, False otherwise.
        """
        if not self.input_ids:
            return True
        for key, value in kwargs.items():  # Iterate over items of kwargs dictionary
            if key not in self.input_ids or self.input_ids[key] != value:
                return False
        return True

    def get_sync_dependencies(self, context: AuthedServiceContext) -> list[UID]:
        # Output ids, user code id, job id
        res = []

        res.extend(self.output_id_list)
        res.append(self.user_code_id)
        if self.job_id:
            res.append(self.job_id)

        return res


@instrument
@serializable(canonical_name="OutputStash", version=1)
class OutputStash(BaseUIDStoreStash):
    object_type = ExecutionOutput
    settings: PartitionSettings = PartitionSettings(
        name=ExecutionOutput.__canonical_name__, object_type=ExecutionOutput
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store)
        self.store = store
        self.settings = self.settings
        self._object_type = self.object_type

    def get_by_user_code_id(
        self, credentials: SyftVerifyKey, user_code_id: UID
    ) -> Result[list[ExecutionOutput], str]:
        qks = QueryKeys(
            qks=[UserCodeIdPartitionKey.with_obj(user_code_id)],
        )
        return self.query_all(
            credentials=credentials, qks=qks, order_by=CreatedAtPartitionKey
        )

    def get_by_job_id(
        self, credentials: SyftVerifyKey, user_code_id: UID
    ) -> Result[ExecutionOutput | None, str]:
        qks = QueryKeys(
            qks=[JobIdPartitionKey.with_obj(user_code_id)],
        )
        res = self.query_all(
            credentials=credentials, qks=qks, order_by=CreatedAtPartitionKey
        )
        if res.is_err():
            return res
        else:
            res = res.ok()
            if len(res) == 0:
                return Ok(None)
            elif len(res) > 1:
                return Err(SyftError(message="Too many outputs found"))
            else:
                return Ok(res[0])

    def get_by_output_policy_id(
        self, credentials: SyftVerifyKey, output_policy_id: UID
    ) -> Result[list[ExecutionOutput], str]:
        qks = QueryKeys(
            qks=[OutputPolicyIdPartitionKey.with_obj(output_policy_id)],
        )
        return self.query_all(
            credentials=credentials, qks=qks, order_by=CreatedAtPartitionKey
        )


@instrument
@serializable(canonical_name="OutputService", version=1)
class OutputService(AbstractService):
    store: DocumentStore
    stash: OutputStash

    def __init__(self, store: DocumentStore):
        self.store = store
        self.stash = OutputStash(store=store)

    @service_method(
        path="output.create",
        name="create",
        roles=GUEST_ROLE_LEVEL,
    )
    def create(
        self,
        context: AuthedServiceContext,
        user_code_id: UID,
        output_ids: UID | list[UID] | dict[str, UID],
        executing_user_verify_key: SyftVerifyKey,
        job_id: UID | None = None,
        output_policy_id: UID | None = None,
        input_ids: dict[str, UID] | None = None,
    ) -> ExecutionOutput | SyftError:
        output = ExecutionOutput.from_ids(
            output_ids=output_ids,
            user_code_id=user_code_id,
            executing_user_verify_key=executing_user_verify_key,
            server_uid=context.server.id,  # type: ignore
            job_id=job_id,
            output_policy_id=output_policy_id,
            input_ids=input_ids,
        )

        res = self.stash.set(context.credentials, output)
        return res

    @service_method(
        path="output.get_by_user_code_id",
        name="get_by_user_code_id",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_by_user_code_id(
        self, context: AuthedServiceContext, user_code_id: UID
    ) -> list[ExecutionOutput] | SyftError:
        result = self.stash.get_by_user_code_id(
            credentials=context.server.verify_key,  # type: ignore
            user_code_id=user_code_id,
        )
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(
        path="output.has_output_read_permissions",
        name="has_output_read_permissions",
        roles=GUEST_ROLE_LEVEL,
    )
    def has_output_read_permissions(
        self,
        context: AuthedServiceContext,
        user_code_id: UID,
        user_verify_key: SyftVerifyKey,
    ) -> bool | SyftError:
        action_service = context.server.get_service("actionservice")
        all_outputs = self.get_by_user_code_id(context, user_code_id)
        if isinstance(all_outputs, SyftError):
            return all_outputs
        for output in all_outputs:
            # TODO tech debt: unclear why code owner can see outputhistory without permissions.
            # It is not a security issue (output history has no data) it is confusing for user
            # if not self.stash.has_permission(
            #     ActionObjectREAD(uid=output.id, credentials=user_verify_key)
            # ):
            #     continue

            # Check if all output ActionObjects have permissions
            result_ids = output.output_id_list
            permissions = [
                ActionObjectREAD(uid=_id.id, credentials=user_verify_key)
                for _id in result_ids
            ]
            if action_service.store.has_permissions(permissions):
                return True

        return False

    @service_method(
        path="output.get_by_job_id",
        name="get_by_job_id",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_by_job_id(
        self, context: AuthedServiceContext, user_code_id: UID
    ) -> ExecutionOutput | None | SyftError:
        result = self.stash.get_by_job_id(
            credentials=context.server.verify_key,  # type: ignore
            user_code_id=user_code_id,
        )
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(
        path="output.get_by_output_policy_id",
        name="get_by_output_policy_id",
        roles=GUEST_ROLE_LEVEL,
    )
    def get_by_output_policy_id(
        self, context: AuthedServiceContext, output_policy_id: UID
    ) -> list[ExecutionOutput] | SyftError:
        result = self.stash.get_by_output_policy_id(
            credentials=context.server.verify_key,  # type: ignore
            output_policy_id=output_policy_id,  # type: ignore
        )
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(
        path="output.get",
        name="get",
        roles=GUEST_ROLE_LEVEL,
    )
    def get(
        self, context: AuthedServiceContext, id: UID
    ) -> ExecutionOutput | SyftError:
        result = self.stash.get_by_uid(context.credentials, id)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(path="output.get_all", name="get_all", roles=GUEST_ROLE_LEVEL)
    def get_all(
        self, context: AuthedServiceContext
    ) -> list[ExecutionOutput] | SyftError:
        result = self.stash.get_all(context.credentials)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())


TYPE_TO_SERVICE[ExecutionOutput] = OutputService



# third party
from IPython.display import HTML
from IPython.display import display

# relative
from ..serde.serializable import serializable
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftObject


@serializable()
class MarkdownDescription(SyftObject):
    # version
    __canonical_name__ = "MarkdownDescription"
    __version__ = SYFT_OBJECT_VERSION_1

    text: str

    def _repr_markdown_(self, wrap_as_python: bool = True, indent: int = 0) -> str:
        style = """
        <style>
            .jp-RenderedHTMLCommon pre {
                background-color: #282c34 !important;
                padding: 10px 10px 10px;
            }
            .jp-RenderedHTMLCommon pre code {
                background-color: #282c34 !important;  /* Set the background color for the text in the code block */
                color: #abb2bf !important;  /* Set text color */
            }
        </style>
        """
        display(HTML(style))
        return self.text


@serializable()
class HTMLObject(SyftObject):
    # version
    __canonical_name__ = "HTMLObject"
    __version__ = SYFT_OBJECT_VERSION_1

    text: str

    def _repr_html_(self) -> str:
        return self.text



"""This folder contains message format for captian proto serialization.

Note: Each capnp message format should have unique hex identifier
(ex: @0xcd0709e35fffa8d8)
These can be generated in terminal by the command `capnp id` after pycapnp installation.
"""



# future
from __future__ import annotations

# stdlib
from typing import TYPE_CHECKING

# relative
from ..abstract_server import ServerSideType
from ..serde.serializable import serializable
from ..service.metadata.server_metadata import ServerMetadataJSON
from ..service.network.routes import ServerRouteType
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftObject
from ..util.assets import load_png_base64
from ..util.notebook_ui.styles import FONT_CSS
from .api import APIModule
from .client import SyftClient
from .client import login
from .client import login_as_guest
from .protocol import SyftProtocol

if TYPE_CHECKING:
    # relative
    from ..orchestra import ServerHandle


@serializable()
class EnclaveMetadata(SyftObject):
    __canonical_name__ = "EnclaveMetadata"
    __version__ = SYFT_OBJECT_VERSION_1

    route: ServerRouteType


@serializable(canonical_name="EnclaveClient", version=1)
class EnclaveClient(SyftClient):
    # TODO: add widget repr for enclave client

    __api_patched = False

    @property
    def code(self) -> APIModule | None:
        if self.api.has_service("code"):
            res = self.api.services.code
            # the order is important here
            # its also important that patching only happens once
            if not self.__api_patched:
                self._request_code_execution = res.request_code_execution
                self.__api_patched = True
            res.request_code_execution = self.request_code_execution
            return res
        return None

    @property
    def requests(self) -> APIModule | None:
        if self.api.has_service("request"):
            return self.api.services.request
        return None

    def connect_to_gateway(
        self,
        via_client: SyftClient | None = None,
        url: str | None = None,
        port: int | None = None,
        handle: ServerHandle | None = None,  # noqa: F821
        email: str | None = None,
        password: str | None = None,
        protocol: str | SyftProtocol = SyftProtocol.HTTP,
    ) -> SyftSuccess | SyftError | None:
        if isinstance(protocol, str):
            protocol = SyftProtocol(protocol)

        if via_client is not None:
            client = via_client
        elif handle is not None:
            client = handle.client
        else:
            client = (
                login_as_guest(url=url, port=port)
                if email is None
                else login(url=url, port=port, email=email, password=password)
            )
            if isinstance(client, SyftError):
                return client

        self.metadata: ServerMetadataJSON = self.metadata
        res = self.exchange_route(client, protocol=protocol)

        if isinstance(res, SyftSuccess):
            if self.metadata:
                return SyftSuccess(
                    message=(
                        f"Connected {self.metadata.server_type} "
                        f"'{self.metadata.name}' to gateway '{client.name}'. "
                        f"{res.message}"
                    )
                )
            else:
                return SyftSuccess(message=f"Connected to '{client.name}' gateway")

        return res

    def get_enclave_metadata(self) -> EnclaveMetadata:
        return EnclaveMetadata(route=self.connection.route)

    def _repr_html_(self) -> str:
        commands = """
        <li><span class='syft-code-block'>&lt;your_client&gt;
        .request_code_execution</span> - submit code to enclave for execution</li>
        """

        command_list = f"""
        <ul style='padding-left: 1em;'>
            {commands}
        </ul>
        """

        small_server_symbol_logo = load_png_base64("small-syft-symbol-logo.png")

        url = getattr(self.connection, "url", None)
        server_details = f"<strong>URL:</strong> {url}<br />" if url else ""
        if self.metadata:
            server_details += f"<strong>Server Type:</strong> {self.metadata.server_type.capitalize()}<br />"
            server_side_type = (
                "Low Side"
                if self.metadata.server_side_type == ServerSideType.LOW_SIDE.value
                else "High Side"
            )
            server_details += (
                f"<strong>Server Side Type:</strong> {server_side_type}<br />"
            )
            server_details += (
                f"<strong>Syft Version:</strong> {self.metadata.syft_version}<br />"
            )

        return f"""
        <style>
            {FONT_CSS}

            .syft-container {{
                padding: 5px;
                font-family: 'Open Sans';
            }}
            .syft-alert-info {{
                color: #1F567A;
                background-color: #C2DEF0;
                border-radius: 4px;
                padding: 5px;
                padding: 13px 10px
            }}
            .syft-code-block {{
                background-color: #f7f7f7;
                border: 1px solid #cfcfcf;
                padding: 0px 2px;
            }}
            .syft-space {{
                margin-top: 1em;
            }}
        </style>
        <div class="syft-client syft-container">
            <img src="{small_server_symbol_logo}" alt="Logo"
            style="width:48px;height:48px;padding:3px;">
            <h2>Welcome to {self.name}</h2>
            <div class="syft-space">
                {server_details}
            </div>
            <div class='syft-alert-info syft-space'>
                &#9432;&nbsp;
                This server is run by the library PySyft to learn more about how it works visit
                <a href="https://github.com/OpenMined/PySyft">github.com/OpenMined/PySyft</a>.
            </div>
            <h4>Commands to Get Started</h4>
            {command_list}
        </div><br />
        """



# stdlib
from datetime import datetime
from datetime import timedelta
from functools import total_ordering
import re
from typing import Any

# third party
from typing_extensions import Self

# relative
from ..serde.serializable import serializable
from .syft_object import SYFT_OBJECT_VERSION_1
from .syft_object import SyftObject
from .uid import UID

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
DATETIME_REGEX = r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}"


def str_is_datetime(str_: str) -> bool:
    return bool(re.match(DATETIME_REGEX, str_))


@serializable()
@total_ordering
class DateTime(SyftObject):
    __canonical_name__ = "DateTime"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID | None = None  # type: ignore
    utc_timestamp: float

    @classmethod
    def now(cls) -> Self:
        return cls(utc_timestamp=datetime.utcnow().timestamp())

    @classmethod
    def from_str(cls, datetime_str: str) -> "DateTime":
        dt = datetime.strptime(datetime_str, DATETIME_FORMAT)
        return cls(utc_timestamp=dt.timestamp())

    def __str__(self) -> str:
        utc_datetime = datetime.utcfromtimestamp(self.utc_timestamp)
        return utc_datetime.strftime(DATETIME_FORMAT)

    def __hash__(self) -> int:
        return hash(self.utc_timestamp)

    def __sub__(self, other: "DateTime") -> "DateTime":
        res = self.utc_timestamp - other.utc_timestamp
        return DateTime(utc_timestamp=res)

    def __eq__(self, other: Any) -> bool:
        if other is None:
            return False
        return self.utc_timestamp == other.utc_timestamp

    def __lt__(self, other: Self) -> bool:
        return self.utc_timestamp < other.utc_timestamp

    def timedelta(self, other: "DateTime") -> timedelta:
        utc_timestamp_delta = self.utc_timestamp - other.utc_timestamp
        return timedelta(seconds=utc_timestamp_delta)


def format_timedelta(local_timedelta: timedelta) -> str:
    total_seconds = int(local_timedelta.total_seconds())
    hours, leftover = divmod(total_seconds, 3600)
    minutes, seconds = divmod(leftover, 60)

    hours_string = f"{hours}:" if hours != 0 else ""
    minutes_string = f"{minutes}:".zfill(3)
    seconds_string = f"{seconds}".zfill(2)

    return f"{hours_string}{minutes_string}{seconds_string}"


def format_timedelta_human_readable(local_timedelta: timedelta) -> str:
    # Returns a human-readable string representing the timedelta
    units = [("day", 86400), ("hour", 3600), ("minute", 60), ("second", 1)]
    total_seconds = int(local_timedelta.total_seconds())

    for unit_name, unit_seconds in units:
        unit_value, total_seconds = divmod(total_seconds, unit_seconds)
        if unit_value > 0:
            if unit_value == 1:
                return f"{unit_value} {unit_name}"
            else:
                return f"{unit_value} {unit_name}s"
    return "0 seconds"



# stdlib
import base64
from collections.abc import Iterable
from enum import Enum
from functools import cache
import json
import os

# third party
import kr8s
from kr8s.objects import APIObject
from kr8s.objects import ConfigMap
from kr8s.objects import Pod
from kr8s.objects import Secret
from kr8s.objects import Service
from pydantic import BaseModel
from typing_extensions import Self

# Kubernetes namespace
KUBERNETES_NAMESPACE = os.getenv("K8S_NAMESPACE", "syft")

# Kubernetes runtime flag
IN_KUBERNETES = os.getenv("CONTAINER_HOST") == "k8s"

# skip pushing to internal registry
USE_INTERNAL_REGISTRY = os.getenv("USE_INTERNAL_REGISTRY", "true").lower() == "true"

# Internal registry URL
DEFAULT_INTERNAL_REGISTRY = f"registry.{KUBERNETES_NAMESPACE}.svc.cluster.local"
INTERNAL_REGISTRY_HOST = os.getenv("INTERNAL_REGISTRY_HOST", DEFAULT_INTERNAL_REGISTRY)


class PodPhase(Enum):
    Pending = "Pending"
    Running = "Running"
    Succeeded = "Succeeded"
    Failed = "Failed"
    Unknown = "Unknown"


class PodCondition(BaseModel):
    pod_scheduled: bool
    containers_ready: bool
    initialized: bool
    ready: bool

    @classmethod
    def from_conditions(cls, conditions: list) -> Self:
        pod_cond = KubeUtils.list_dict_unpack(conditions, key="type", value="status")
        pod_cond_flags = {k: v == "True" for k, v in pod_cond.items()}
        return cls(
            pod_scheduled=pod_cond_flags.get("PodScheduled", False),
            containers_ready=pod_cond_flags.get("ContainersReady", False),
            initialized=pod_cond_flags.get("Initialized", False),
            ready=pod_cond_flags.get("Ready", False),
        )


class ContainerStatus(BaseModel):
    ready: bool
    running: bool
    waiting: bool
    reason: str | None = None  # when waiting=True
    message: str | None = None  # when waiting=True
    startedAt: str | None = None  # when running=True

    @classmethod
    def from_status(cls, cstatus: dict) -> Self:
        cstate = cstatus.get("state", {})

        return cls(
            ready=cstatus.get("ready", False),
            running="running" in cstate,
            waiting="waiting" in cstate,
            reason=cstate.get("waiting", {}).get("reason", None),
            message=cstate.get("waiting", {}).get("message", None),
            startedAt=cstate.get("running", {}).get("startedAt", None),
        )


class PodStatus(BaseModel):
    phase: PodPhase
    condition: PodCondition
    container: ContainerStatus

    @classmethod
    def from_status_dict(cls, status: dict) -> Self:
        return cls(
            phase=PodPhase(status.get("phase", "Unknown")),
            condition=PodCondition.from_conditions(status.get("conditions", [])),
            container=ContainerStatus.from_status(
                status.get("containerStatuses", {})[0]
            ),
        )


@cache
def get_kr8s_client() -> kr8s.Api:
    if not IN_KUBERNETES:
        raise RuntimeError("Not inside a kubernetes environment")
    return kr8s.api(namespace=KUBERNETES_NAMESPACE)


class KubeUtils:
    """
    This class contains utility functions for interacting with kubernetes objects.

    DO NOT call `get_kr8s_client()` inside this class, instead pass it as an argument to the functions.
    This is to avoid calling these functions on resources across namespaces!
    """

    @staticmethod
    def resolve_pod(client: kr8s.Api, pod: str | Pod) -> Pod | None:
        """Return the first pod that matches the given name"""
        if isinstance(pod, Pod):
            return pod

        for _pod in client.get("pods", pod):
            return _pod

        return None

    @staticmethod
    def get_logs(pods: list[Pod]) -> str:
        """Combine and return logs for all the pods as a single string."""
        return "\n".join(
            f"----------Logs for pod={pod.metadata.name}----------\n{''.join(pod.logs())}"
            for pod in pods
        )

    @staticmethod
    def get_pod_status(pod: Pod) -> PodStatus | None:
        """Map the status of the given pod to PodStatuss."""
        if not pod:
            return None
        return PodStatus.from_status_dict(pod.status)

    @staticmethod
    def get_pod_env(pod: Pod) -> list[dict] | None:
        """Return the environment variables of the first container in the pod."""
        if not pod:
            return None

        for container in pod.spec.containers:
            return container.env.to_list()

        return None

    @staticmethod
    def get_container_exit_code(pods: list[Pod]) -> list[int]:
        """Return the exit codes of all the containers in the given pods."""
        return [
            container_status.state.terminated.exitCode
            for pod in pods
            for container_status in pod.status.containerStatuses
        ]

    @staticmethod
    def get_container_exit_message(pods: list[Pod]) -> str | None:
        """Return the exit message of the first container that exited with non-zero code."""
        for pod in pods:
            for container_status in pod.status.containerStatuses:
                if container_status.state.terminated.exitCode != 0:
                    continue
                return container_status.state.terminated.message
        return None

    @staticmethod
    def b64encode_secret(data: str) -> str:
        """Convert the data to base64 encoded string for Secret."""
        return base64.b64encode(data.encode()).decode()

    @staticmethod
    def get_configmap(client: kr8s.Api, name: str) -> ConfigMap | None:
        config_map = client.get("configmaps", name)
        return config_map[0] if config_map else None

    @staticmethod
    def get_service(client: kr8s.Api, name: str) -> Service | None:
        service = client.get("services", name)
        return service[0] if service else None

    @staticmethod
    def update_configmap(
        config_map: ConfigMap,
        patch: dict,
    ) -> None:
        existing_data = config_map.raw
        existing_data.update(patch)
        config_map.patch(patch=existing_data)

    @staticmethod
    def create_dockerconfig_secret(
        secret_name: str,
        component: str,
        registries: Iterable[tuple[str, str, str]],
    ) -> Secret:
        auths = {}

        for url, uname, passwd in registries:
            auths[url] = {
                "username": uname,
                "password": passwd,
                "auth": KubeUtils.b64encode_secret(f"{uname}:{passwd}"),
            }

        config_str = json.dumps({"auths": auths})

        return KubeUtils.create_secret(
            secret_name=secret_name,
            type="kubernetes.io/dockerconfigjson",
            component=component,
            data={
                ".dockerconfigjson": KubeUtils.b64encode_secret(config_str),
            },
        )

    @staticmethod
    def create_secret(
        secret_name: str,
        type: str,
        component: str,
        data: str,
        encoded: bool = True,
    ) -> Secret:
        if not encoded:
            for k, v in data.items():
                data[k] = KubeUtils.b64encode_secret(v)  # type: ignore

        secret = Secret(
            {
                "metadata": {
                    "name": secret_name,
                    "labels": {
                        "app.kubernetes.io/name": KUBERNETES_NAMESPACE,
                        "app.kubernetes.io/component": component,
                        "app.kubernetes.io/managed-by": "kr8s",
                    },
                },
                "type": type,
                "data": data,
            }
        )
        return KubeUtils.create_or_get(secret)

    @staticmethod
    def create_or_get(obj: APIObject) -> APIObject:
        if obj.exists():
            obj.refresh()
        else:
            obj.create()
        return obj

    @staticmethod
    def patch_env_vars(env_list: list[dict], env_dict: dict) -> list[dict]:
        """Patch kubernetes pod environment variables in the list with the provided dictionary."""

        # update existing
        for item in env_list:
            k = item["name"]
            if k in env_dict:
                v = env_dict.pop(k)
                item["value"] = v

        # append remaining
        for k, v in env_dict.items():
            env_list.append({"name": k, "value": v})

        return env_list

    @staticmethod
    def list_dict_unpack(
        input_list: list[dict],
        key: str = "key",
        value: str = "value",
    ) -> dict:
        # Snapshot from kr8s._data_utils
        return {i[key]: i[value] for i in input_list}
